import json
import logging
import random
from typing import Any, Dict, Optional
import time
import requests

from desktop_env.actions import KEYBOARD_KEYS

logger = logging.getLogger("desktopenv.pycontroller")


class PythonController:
    def __init__(self, vm_ip: str,
                 server_port: int,
                 pkgs_prefix: str = "import pyautogui; import time; pyautogui.FAILSAFE = False; {command}"):
        self.vm_ip = vm_ip
        self.http_server = f"http://{vm_ip}:{server_port}"
        self.pkgs_prefix = pkgs_prefix  # fixme: this is a hacky way to execute python commands. fix it and combine it with installation of packages
        self.retry_times = 3
        self.retry_interval = 5

    def get_screenshot(self) -> Optional[bytes]:
        """
        Gets a screenshot from the server. With the cursor. None -> no screenshot or unexpected error.
        """

        for _ in range(self.retry_times):
            try:
                response = requests.get(self.http_server + "/screenshot")
                if response.status_code == 200:
                    logger.info("Got screenshot successfully")
                    return response.content
                else:
                    logger.error("Failed to get screenshot. Status code: %d", response.status_code)
                    logger.info("Retrying to get screenshot.")
            except Exception as e:
                logger.error("An error occurred while trying to get the screenshot: %s", e)
                logger.info("Retrying to get screenshot.")
            time.sleep(self.retry_interval)

        logger.error("Failed to get screenshot.")
        return None

    def get_accessibility_tree(self) -> Optional[str]:
        """
        Gets the accessibility tree from the server. None -> no accessibility tree or unexpected error.
        """

        for _ in range(self.retry_times):
            try:
                response: requests.Response = requests.get(self.http_server + "/accessibility")
                if response.status_code == 200:
                    logger.info("Got accessibility tree successfully")
                    return response.json()["AT"]
                else:
                    logger.error("Failed to get accessibility tree. Status code: %d", response.status_code)
                    logger.info("Retrying to get accessibility tree.")
            except Exception as e:
                logger.error("An error occurred while trying to get the accessibility tree: %s", e)
                logger.info("Retrying to get accessibility tree.")
            time.sleep(self.retry_interval)

        logger.error("Failed to get accessibility tree.")
        return None

    def get_terminal_output(self) -> Optional[str]:
        """
        Gets the terminal output from the server. None -> no terminal output or unexpected error.
        """

        for _ in range(self.retry_times):
            try:
                response = requests.get(self.http_server + "/terminal")
                if response.status_code == 200:
                    logger.info("Got terminal output successfully")
                    return response.json()["output"]
                else:
                    logger.error("Failed to get terminal output. Status code: %d", response.status_code)
                    logger.info("Retrying to get terminal output.")
            except Exception as e:
                logger.error("An error occurred while trying to get the terminal output: %s", e)
                logger.info("Retrying to get terminal output.")
            time.sleep(self.retry_interval)

        logger.error("Failed to get terminal output.")
        return None

    def get_file(self, file_path: str) -> Optional[bytes]:
        """
        Gets a file from the server.
        """

        for _ in range(self.retry_times):
            try:
                response = requests.post(self.http_server + "/file", data={"file_path": file_path})
                if response.status_code == 200:
                    logger.info("File downloaded successfully")
                    return response.content
                else:
                    logger.error("Failed to get file. Status code: %d", response.status_code)
                    logger.info("Retrying to get file.")
            except Exception as e:
                logger.error("An error occurred while trying to get the file: %s", e)
                logger.info("Retrying to get file.")
            time.sleep(self.retry_interval)

        logger.error("Failed to get file.")
        return None

    def execute_python_command(self, command: str) -> None:
        """
        Executes a python command on the server.
        It can be used to execute the pyautogui commands, or... any other python command. who knows?
        """
        # command_list = ["python", "-c", self.pkgs_prefix.format(command=command)]
        command_list = ["python", "-c", self.pkgs_prefix.format(command=command)]
        payload = json.dumps({"command": command_list, "shell": False})

        for _ in range(self.retry_times):
            try:
                response = requests.post(self.http_server + "/execute", headers={'Content-Type': 'application/json'},
                                         data=payload, timeout=90)
                if response.status_code == 200:
                    logger.info("Command executed successfully: %s", response.text)
                    return response.json()
                else:
                    logger.error("Failed to execute command. Status code: %d", response.status_code)
                    logger.info("Retrying to execute command.")
            except requests.exceptions.ReadTimeout:
                break
            except Exception as e:
                logger.error("An error occurred while trying to execute the command: %s", e)
                logger.info("Retrying to execute command.")
            time.sleep(self.retry_interval)

        logger.error("Failed to execute command.")
        return None

    def execute_action(self, action: Dict[str, Any]):
        """
        Executes an action on the server computer.
        """
        if action in ['WAIT', 'FAIL', 'DONE']:
            return

        action_type = action["action_type"]
        parameters = action["parameters"] if "parameters" in action else {param: action[param] for param in action if param != 'action_type'}
        move_mode = random.choice(
            ["pyautogui.easeInQuad", "pyautogui.easeOutQuad", "pyautogui.easeInOutQuad", "pyautogui.easeInBounce",
             "pyautogui.easeInElastic"])
        duration = random.uniform(0.5, 1)

        if action_type == "MOVE_TO":
            if parameters == {} or None:
                self.execute_python_command("pyautogui.moveTo()")
            elif "x" in parameters and "y" in parameters:
                x = parameters["x"]
                y = parameters["y"]
                self.execute_python_command(f"pyautogui.moveTo({x}, {y}, {duration}, {move_mode})")
            else:
                raise Exception(f"Unknown parameters: {parameters}")

        elif action_type == "CLICK":
            if parameters == {} or None:
                self.execute_python_command("pyautogui.click()")
            elif "button" in parameters and "x" in parameters and "y" in parameters:
                button = parameters["button"]
                x = parameters["x"]
                y = parameters["y"]
                if "num_clicks" in parameters:
                    num_clicks = parameters["num_clicks"]
                    self.execute_python_command(
                        f"pyautogui.click(button='{button}', x={x}, y={y}, clicks={num_clicks})")
                else:
                    self.execute_python_command(f"pyautogui.click(button='{button}', x={x}, y={y})")
            elif "button" in parameters and "x" not in parameters and "y" not in parameters:
                button = parameters["button"]
                if "num_clicks" in parameters:
                    num_clicks = parameters["num_clicks"]
                    self.execute_python_command(f"pyautogui.click(button='{button}', clicks={num_clicks})")
                else:
                    self.execute_python_command(f"pyautogui.click(button='{button}')")
            elif "button" not in parameters and "x" in parameters and "y" in parameters:
                x = parameters["x"]
                y = parameters["y"]
                if "num_clicks" in parameters:
                    num_clicks = parameters["num_clicks"]
                    self.execute_python_command(f"pyautogui.click(x={x}, y={y}, clicks={num_clicks})")
                else:
                    self.execute_python_command(f"pyautogui.click(x={x}, y={y})")
            else:
                raise Exception(f"Unknown parameters: {parameters}")

        elif action_type == "MOUSE_DOWN":
            if parameters == {} or None:
                self.execute_python_command("pyautogui.mouseDown()")
            elif "button" in parameters:
                button = parameters["button"]
                self.execute_python_command(f"pyautogui.mouseDown(button='{button}')")
            else:
                raise Exception(f"Unknown parameters: {parameters}")

        elif action_type == "MOUSE_UP":
            if parameters == {} or None:
                self.execute_python_command("pyautogui.mouseUp()")
            elif "button" in parameters:
                button = parameters["button"]
                self.execute_python_command(f"pyautogui.mouseUp(button='{button}')")
            else:
                raise Exception(f"Unknown parameters: {parameters}")

        elif action_type == "RIGHT_CLICK":
            if parameters == {} or None:
                self.execute_python_command("pyautogui.rightClick()")
            elif "x" in parameters and "y" in parameters:
                x = parameters["x"]
                y = parameters["y"]
                self.execute_python_command(f"pyautogui.rightClick(x={x}, y={y})")
            else:
                raise Exception(f"Unknown parameters: {parameters}")

        elif action_type == "DOUBLE_CLICK":
            if parameters == {} or None:
                self.execute_python_command("pyautogui.doubleClick()")
            elif "x" in parameters and "y" in parameters:
                x = parameters["x"]
                y = parameters["y"]
                self.execute_python_command(f"pyautogui.doubleClick(x={x}, y={y})")
            else:
                raise Exception(f"Unknown parameters: {parameters}")

        elif action_type == "DRAG_TO":
            if "x" in parameters and "y" in parameters:
                x = parameters["x"]
                y = parameters["y"]
                self.execute_python_command(
                    f"pyautogui.dragTo({x}, {y}, duration=1.0, button='left', mouseDownUp=True)")

        elif action_type == "SCROLL":
            # todo: check if it is related to the operating system, as https://github.com/TheDuckAI/DuckTrack/blob/main/ducktrack/playback.py pointed out
            if "dx" in parameters and "dy" in parameters:
                dx = parameters["dx"]
                dy = parameters["dy"]
                self.execute_python_command(f"pyautogui.hscroll({dx})")
                self.execute_python_command(f"pyautogui.vscroll({dy})")
            elif "dx" in parameters and "dy" not in parameters:
                dx = parameters["dx"]
                self.execute_python_command(f"pyautogui.hscroll({dx})")
            elif "dx" not in parameters and "dy" in parameters:
                dy = parameters["dy"]
                self.execute_python_command(f"pyautogui.vscroll({dy})")
            else:
                raise Exception(f"Unknown parameters: {parameters}")

        elif action_type == "TYPING":
            if "text" not in parameters:
                raise Exception(f"Unknown parameters: {parameters}")
            # deal with special ' and \ characters
            # text = parameters["text"].replace("\\", "\\\\").replace("'", "\\'")
            # self.execute_python_command(f"pyautogui.typewrite('{text}')")
            text = parameters["text"]
            self.execute_python_command("pyautogui.typewrite({:})".format(repr(text)))

        elif action_type == "PRESS":
            if "key" not in parameters:
                raise Exception(f"Unknown parameters: {parameters}")
            key = parameters["key"]
            if key.lower() not in KEYBOARD_KEYS:
                raise Exception(f"Key must be one of {KEYBOARD_KEYS}")
            self.execute_python_command(f"pyautogui.press('{key}')")

        elif action_type == "KEY_DOWN":
            if "key" not in parameters:
                raise Exception(f"Unknown parameters: {parameters}")
            key = parameters["key"]
            if key.lower() not in KEYBOARD_KEYS:
                raise Exception(f"Key must be one of {KEYBOARD_KEYS}")
            self.execute_python_command(f"pyautogui.keyDown('{key}')")

        elif action_type == "KEY_UP":
            if "key" not in parameters:
                raise Exception(f"Unknown parameters: {parameters}")
            key = parameters["key"]
            if key.lower() not in KEYBOARD_KEYS:
                raise Exception(f"Key must be one of {KEYBOARD_KEYS}")
            self.execute_python_command(f"pyautogui.keyUp('{key}')")

        elif action_type == "HOTKEY":
            if "keys" not in parameters:
                raise Exception(f"Unknown parameters: {parameters}")
            keys = parameters["keys"]
            if not isinstance(keys, list):
                raise Exception("Keys must be a list of keys")
            for key in keys:
                if key.lower() not in KEYBOARD_KEYS:
                    raise Exception(f"Key must be one of {KEYBOARD_KEYS}")

            keys_para_rep = "', '".join(keys)
            self.execute_python_command(f"pyautogui.hotkey('{keys_para_rep}')")

        elif action_type in ['WAIT', 'FAIL', 'DONE']:
            pass

        else:
            raise Exception(f"Unknown action type: {action_type}")

    # Record video
    def start_recording(self):
        """
        Starts recording the screen.
        """

        for _ in range(self.retry_times):
            try:
                response = requests.post(self.http_server + "/start_recording")
                if response.status_code == 200:
                    logger.info("Recording started successfully")
                    return
                else:
                    logger.error("Failed to start recording. Status code: %d", response.status_code)
                    logger.info("Retrying to start recording.")
            except Exception as e:
                logger.error("An error occurred while trying to start recording: %s", e)
                logger.info("Retrying to start recording.")
            time.sleep(self.retry_interval)

        logger.error("Failed to start recording.")

    def end_recording(self, dest: str):
        """
        Ends recording the screen.
        """

        for _ in range(self.retry_times):
            try:
                response = requests.post(self.http_server + "/end_recording")
                if response.status_code == 200:
                    logger.info("Recording stopped successfully")
                    with open(dest, 'wb') as f:
                        for chunk in response.iter_content(chunk_size=8192):
                            if chunk:
                                f.write(chunk)
                    return
                else:
                    logger.error("Failed to stop recording. Status code: %d", response.status_code)
                    logger.info("Retrying to stop recording.")
            except Exception as e:
                logger.error("An error occurred while trying to stop recording: %s", e)
                logger.info("Retrying to stop recording.")
            time.sleep(self.retry_interval)

        logger.error("Failed to stop recording.")

    # Additional info
    def get_vm_platform(self):
        """
        Gets the size of the vm screen.
        """
        return self.execute_python_command("import platform; print(platform.system())")['output'].strip()

    def get_vm_screen_size(self):
        """
        Gets the size of the vm screen.
        """

        for _ in range(self.retry_times):
            try:
                response = requests.post(self.http_server + "/screen_size")
                if response.status_code == 200:
                    logger.info("Got screen size successfully")
                    return response.json()
                else:
                    logger.error("Failed to get screen size. Status code: %d", response.status_code)
                    logger.info("Retrying to get screen size.")
            except Exception as e:
                logger.error("An error occurred while trying to get the screen size: %s", e)
                logger.info("Retrying to get screen size.")
            time.sleep(self.retry_interval)

        logger.error("Failed to get screen size.")
        return None

    def get_vm_window_size(self, app_class_name: str):
        """
        Gets the size of the vm app window.
        """

        for _ in range(self.retry_times):
            try:
                response = requests.post(self.http_server + "/window_size", data={"app_class_name": app_class_name})
                if response.status_code == 200:
                    logger.info("Got window size successfully")
                    return response.json()
                else:
                    logger.error("Failed to get window size. Status code: %d", response.status_code)
                    logger.info("Retrying to get window size.")
            except Exception as e:
                logger.error("An error occurred while trying to get the window size: %s", e)
                logger.info("Retrying to get window size.")
            time.sleep(self.retry_interval)

        logger.error("Failed to get window size.")
        return None

    def get_vm_wallpaper(self):
        """
        Gets the wallpaper of the vm.
        """

        for _ in range(self.retry_times):
            try:
                response = requests.post(self.http_server + "/wallpaper")
                if response.status_code == 200:
                    logger.info("Got wallpaper successfully")
                    return response.content
                else:
                    logger.error("Failed to get wallpaper. Status code: %d", response.status_code)
                    logger.info("Retrying to get wallpaper.")
            except Exception as e:
                logger.error("An error occurred while trying to get the wallpaper: %s", e)
                logger.info("Retrying to get wallpaper.")
            time.sleep(self.retry_interval)

        logger.error("Failed to get wallpaper.")
        return None

    def get_vm_desktop_path(self) -> Optional[str]:
        """
        Gets the desktop path of the vm.
        """

        for _ in range(self.retry_times):
            try:
                response = requests.post(self.http_server + "/desktop_path")
                if response.status_code == 200:
                    logger.info("Got desktop path successfully")
                    return response.json()["desktop_path"]
                else:
                    logger.error("Failed to get desktop path. Status code: %d", response.status_code)
                    logger.info("Retrying to get desktop path.")
            except Exception as e:
                logger.error("An error occurred while trying to get the desktop path: %s", e)
                logger.info("Retrying to get desktop path.")
            time.sleep(self.retry_interval)

        logger.error("Failed to get desktop path.")
        return None

    def get_vm_directory_tree(self, path) -> Optional[Dict[str, Any]]:
        """
        Gets the directory tree of the vm.
        """
        payload = json.dumps({"path": path})

        for _ in range(self.retry_times):
            try:
                response = requests.post(self.http_server + "/list_directory", headers={'Content-Type': 'application/json'}, data=payload)
                if response.status_code == 200:
                    logger.info("Got directory tree successfully")
                    return response.json()["directory_tree"]
                else:
                    logger.error("Failed to get directory tree. Status code: %d", response.status_code)
                    logger.info("Retrying to get directory tree.")
            except Exception as e:
                logger.error("An error occurred while trying to get directory tree: %s", e)
                logger.info("Retrying to get directory tree.")
            time.sleep(self.retry_interval)

        logger.error("Failed to get directory tree.")
        return None



import json
import logging
import os
import platform
import textwrap
from typing import Dict, List, Optional, Tuple

from gui_agents.agents.worker import Worker
from gui_agents.agents.manager import Manager
from gui_agents.agents.grounding import Grounding, FastGrounding
from gui_agents.utils.common_utils import Node
from gui_agents.agents.global_state import GlobalState
from gui_agents.store.registry import Registry
from gui_agents.utils.common_utils import (
    parse_single_code_from_string,
    sanitize_code,
    extract_first_agent_function,
    agent_log_to_string,
)
from gui_agents.tools.tools import Tools
from gui_agents.agents.stream_manager import stream_manager

logger = logging.getLogger("desktopenv.agent")

def load_config():
    """
    Load tool configurations from the repository's tools/tools_config.json and produce a mapping keyed by tool name.
    
    Returns:
        tuple: (tools_config, tools_dict) where `tools_config` is the parsed JSON object from tools_config.json, and `tools_dict` is a dict mapping each tool's `tool_name` to a dict with `provider` and `model`.
    """
    # Load tools configuration from tools_config.json
    tools_config_path = os.path.join(os.path.dirname(os.path.dirname(__file__)), "tools", "tools_config.json")
    with open(tools_config_path, "r") as f:
        tools_config = json.load(f)
        print(f"Loaded tools configuration from: {tools_config_path}")
        tools_dict = {}
        for tool in tools_config["tools"]:
            tool_name = tool["tool_name"]
            tools_dict[tool_name] = {
                "provider": tool["provider"],
                "model": tool["model_name"]
            }
        print(f"Tools configuration: {tools_dict}")
        return tools_config,tools_dict

class UIAgent:
    """Base class for UI automation agents"""

    def __init__(
        self,
        platform: str = platform.system().lower(),
    ):
        """Initialize UIAgent

        Args:
            platform: Operating system platform (macos, linux, windows)
        """
        self.platform = platform

    def reset(self) -> None:
        """
        Reset the agent to its initial internal state.
        
        Performs any subclass-specific reinitialization needed so the agent is ready to start a new task or episode.
        """
        pass

    def _send_stream_message(self, task_id: str, stage: str, message: str) -> None:
        """
        Safely send stream message to task stream.
        """
        if not task_id:
            return

        stream_manager.add_message_threadsafe(task_id, stage, message)

    def predict(self, instruction: str, observation: Dict) -> Tuple[Dict, List[str]]|None:
        """
        Produce the next agent information and action sequence for the given instruction and current observation.
        
        Returns:
            (info, actions) where `info` is a dictionary containing planner, executor and evaluator metadata (including subtask metadata and statuses) and `actions` is a list of action strings to execute; returns `None` if no prediction is available.
        """
        pass

    def update_narrative_memory(self, trajectory: str) -> None:
        """Update narrative memory with task trajectory

        Args:
            trajectory: String containing task execution trajectory
        """
        pass

    def update_episodic_memory(self, meta_data: Dict, subtask_trajectory: str) -> str|None:
        """Update episodic memory with subtask trajectory

        Args:
            meta_data: Metadata about current subtask execution
            subtask_trajectory: String containing subtask execution trajectory

        Returns:
            Updated subtask trajectory
        """
        pass

class AgentS2(UIAgent):
    """Agent that uses hierarchical planning and directed acyclic graph modeling for UI automation"""

    def __init__(
        self,
        platform: str = platform.system(),
        screen_size: List[int] = [1920, 1080],
        memory_root_path: str = os.getcwd(),
        memory_folder_name: str = "kb_s2",
        kb_release_tag: str = "v0.2.2",
        enable_takeover: bool = False,
        enable_search: bool = True,
        tools_config: dict | None = None,
    ):
        """
        Initialize an AgentS2 instance and prepare its tools and local knowledge base.
        
        If `tools_config` is provided, build `Tools_dict` mapping each `tool_name` to its config (renaming `model_name` to `model` and removing `tool_name`). If `tools_config` is not provided, load configuration via `load_config()`. Ensure a platform-specific knowledge base directory exists under `memory_root_path/memory_folder_name` (creating it if missing). Sets initial attributes (platform, screen_size, memory paths, flags) and initializes internal state via `reset()`.
        
        Parameters:
            tools_config (dict | None): Optional pre-loaded tools configuration; when present it is transformed into `Tools_dict`. Omit to load configuration from disk.
        """
        platform = platform.lower()
        super().__init__(
            platform,
        )

        self.memory_root_path = memory_root_path
        self.memory_folder_name = memory_folder_name
        self.kb_release_tag = kb_release_tag
        self.screen_size = screen_size
        self.enable_takeover = enable_takeover
        self.enable_search = enable_search
        self.task_id = None  # Will be set when task starts

        if tools_config is not None:
            self.tools_config = tools_config
            # Create the dictionary mapping from the list-based config
            self.Tools_dict = {}
            for tool in self.tools_config["tools"]:
                tool_name = tool["tool_name"]
                # Create a copy of the tool's config to avoid modifying the original
                config_copy = tool.copy()
                # Rename 'model_name' to 'model' for consistency in downstream use
                if 'model_name' in config_copy:
                    config_copy['model'] = config_copy.pop('model_name')
                # Remove tool_name as it's now the key
                config_copy.pop('tool_name', None)
                self.Tools_dict[tool_name] = config_copy
        else:
            self.tools_config, self.Tools_dict = load_config()

        # Initialize agent's knowledge base path
        self.local_kb_path = os.path.join(
            self.memory_root_path, self.memory_folder_name
        )

        # Check if knowledge base exists
        kb_platform_path = os.path.join(self.local_kb_path, self.platform)
        if not os.path.exists(kb_platform_path):
            print(f"Warning: Knowledge base for {self.platform} platform not found in {self.local_kb_path}")
            os.makedirs(kb_platform_path, exist_ok=True)
            print(f"Created directory: {kb_platform_path}")
            # raise FileNotFoundError(f"Knowledge base path does not exist: {kb_platform_path}")
        else:
            print(f"Found local knowledge base path: {kb_platform_path}")

    def reset(self) -> None:
        """
        Reinitialize core components and reset the agent's runtime state.
        
        Recreates the Manager, Worker, and Grounding components using the agent's current configuration,
        resets planning/execution flags and counters, clears subtask-related state, reloads the shared
        global state from the registry, and propagates the agent's task_id to the components when present.
        """
        # Initialize core components

        self.manager = Manager(
            Tools_dict=self.Tools_dict,
            local_kb_path=self.local_kb_path,
            platform=self.platform,
            enable_search=self.enable_search,  # Pass global switch to Manager
        )

        self.worker = Worker(
            Tools_dict=self.Tools_dict,
            local_kb_path=self.local_kb_path,
            platform=self.platform,
            enable_takeover=self.enable_takeover,
            enable_search=self.enable_search,  # Pass global switch to Worker
            tools_config=self.tools_config,    # Pass complete tools configuration
        )

        self.grounding = Grounding(
            Tools_dict=self.Tools_dict,
            platform=self.platform,
            width=self.screen_size[0],
            height=self.screen_size[1]
        )

        # Reset state variables
        self.requires_replan: bool = True
        self.needs_next_subtask: bool = True
        self.step_count: int = 0
        self.turn_count: int = 0
        self.failure_subtask: Optional[Node] = None
        self.should_send_action: bool = False
        self.completed_tasks: List[Node] = []
        self.current_subtask: Optional[Node] = None
        self.subtasks: List[Node] = []
        self.search_query: str = ""
        self.subtask_status: str = "Start"
        # Use task-specific registry if task_id is available, otherwise fall back to global registry
        if self.task_id:
            self.global_state: GlobalState = Registry.get_from_context("GlobalStateStore", self.task_id) # type: ignore
        else:
            self.global_state: GlobalState = Registry.get("GlobalStateStore") # type: ignore

        # Pass task_id to components
        self.manager.set_task_id(self.task_id)
        self.worker.set_task_id(self.task_id)
        # Grounding doesn't have task_id in normal mode, but we set it if available
        if hasattr(self, 'grounding') and hasattr(self.grounding, 'set_task_id'):
            self.grounding.set_task_id(self.task_id)

    def set_task_id(self, task_id: str) -> None:
        """
        Set the task identifier and propagate it to internal components used for streaming.
        
        Parameters:
            task_id (str): Identifier for the current task; assigned to this agent and, if present, to its manager and worker so stream messages are tagged consistently.
        """
        self.task_id = task_id
        # Also set task_id for components if they exist
        if hasattr(self, 'manager') and self.manager:
            self.manager.set_task_id(task_id)
        if hasattr(self, 'worker') and self.worker:
            self.worker.set_task_id(task_id)

    def reset_executor_state(self) -> None:
        """Reset executor and step counter"""
        self.worker.reset()
        self.step_count = 0

    def predict(self, instruction: str, observation: Dict) -> Tuple[Dict, List[str]]:
        # Initialize the three info dictionaries
        """
        Produce the next executor actions and diagnostic information for the current task step.

        This method coordinates planning, subtask selection, action generation, grounding (code extraction and execution), and status updates. It may trigger replanning, advance to the next subtask, mark subtasks as completed or failed, and emit stream messages and logs. The returned info merges planner, executor, and evaluator metadata and includes current subtask details.

        Parameters:
            instruction (str): The user or system instruction describing the task to accomplish; forwarded to the manager/worker as the task utterance.
            observation (Dict): Current environment observation/state used for grounding and coordinate assignment.

        Returns:
            info (Dict): A merged dictionary containing planner_info, executor_info, evaluator_info and the keys `subtask`, `subtask_info`, and `subtask_status`.
            actions (List[Dict]): List of action dictionaries produced for execution (may include actions with type "DONE", failure indicators, or other executor-generated actions).
        """
        # Check for cancellation before starting prediction
        if self.global_state.is_cancelled():
            logger.info("AgentS2 prediction cancelled by user request")
            return {
                "subtask": "cancelled",
                "subtask_info": "",
                "subtask_status": "cancelled",
                "reflection": "Task was cancelled",
                "executor_plan": "agent.done()"
            }, ["done"]

        planner_info = {}
        executor_info = {}
        evaluator_info = {
            "obs_evaluator_response": "",
            "num_input_tokens_evaluator": 0,
            "num_output_tokens_evaluator": 0,
            "evaluator_cost": 0.0,
        }
        actions = []

        # 记录预测开始时间
        import time
        predict_start_time = time.time()

        # If the DONE response by the executor is for a subtask, then the agent should continue with the next subtask without sending the action to the environment
        while not self.should_send_action:
            # Check for cancellation in each iteration
            if self.global_state.is_cancelled():
                logger.info("AgentS2 prediction loop cancelled by user request")
                return {
                    "subtask": "cancelled",
                    "subtask_info": "",
                    "subtask_status": "cancelled",
                    "reflection": "Task was cancelled",
                    "executor_plan": "agent.done()"
                }, [{"type": "DONE"}]
            time.sleep(5.0)
            self.subtask_status = "In"
            # Always time get_action_queue, even if not called
            import time
            manager_start = time.time()
            # If replan is true, generate a new plan. True at start, after a failed plan, or after subtask completion
            if self.requires_replan:
                logger.info("(RE)PLANNING...")

                # Stream planning start message
                self._send_stream_message(self.task_id, "planning", f"Start planning task steps (Step {self.step_count + 1})...")

                Manager_info, self.subtasks = self.manager.get_action_queue(
                    Tu=self.global_state.get_Tu(),
                    observation=self.global_state.get_obs_for_manager(),
                    running_state=self.global_state.get_running_state(),
                    failed_subtask=self.failure_subtask,
                    completed_subtasks_list=self.global_state.get_completed_subtasks(),
                    remaining_subtasks_list=self.global_state.get_remaining_subtasks(),
                )
                self.global_state.set_remaining_subtasks(self.subtasks) # type: ignore

                self.requires_replan = False
                if "search_query" in Manager_info:
                    self.search_query = Manager_info["search_query"]
                else:
                    self.search_query = ""

                # Stream planning completion message
                self._send_stream_message(self.task_id, "planning", f"Planning completed, {len(self.subtasks)} subtasks generated")
            get_action_queue_time = time.time() - manager_start
            logger.info(f"[Timing] manager.get_action_queue execution time: {get_action_queue_time:.2f} seconds")
            self.global_state.log_operation(
                module="manager",
                operation="manager.get_action_queue",
                data={"duration": get_action_queue_time}
            )

            # use the exectuor to complete the topmost subtask
            if self.needs_next_subtask:
                logger.info("GETTING NEXT SUBTASK...")

                # this can be empty if the DAG planner deems that all subtasks are completed
                if len(self.subtasks) <= 0:
                    self.requires_replan = True
                    self.needs_next_subtask = True
                    self.failure_subtask = None
                    if self.current_subtask is not None:
                        self.global_state.add_completed_subtask(self.current_subtask)
                    # reset executor state
                    self.reset_executor_state()
                    self.should_send_action = True
                    self.subtask_status = "Done"
                    executor_info = {
                        "executor_plan": "agent.done()",
                        "plan_code": "agent.done()",
                        "reflection": "agent.done()",
                    }
                    actions = [{"type": "DONE"}]

                    # Stream task completion message
                    self._send_stream_message(self.task_id, "completion", "🎉 Mission Completed! All subtasks have been successfully executed")

                    self.global_state.log_operation(
                        module="agent",
                        operation="task_complete",
                        data={
                            "content": "All subtasks completed, task finished",
                            "status": "done"
                        }
                    )
                    break

                self.current_subtask = self.subtasks.pop(0)
                self.global_state.set_remaining_subtasks(self.subtasks)
                logger.info(f"NEXT SUBTASK: {self.current_subtask}")
                logger.info(f"REMAINING SUBTASKS: {self.subtasks}")
                logger.info(f"REMAINING SUBTASKS FROM GLOBAL STATE: {self.global_state.get_remaining_subtasks()}")
                self.needs_next_subtask = False
                self.subtask_status = "Start"

                # Stream current subtask message
                if self.current_subtask is not None:
                    self._send_stream_message(self.task_id, "subtask", f"Start executing subtasks: {self.current_subtask.name}")
                else:
                    self._send_stream_message(self.task_id, "subtask", "Start executing a new subtask")

                self.global_state.log_operation(
                    module="agent",
                    operation="current_subtask",
                    data={
                        "content": str(self.current_subtask) if self.current_subtask is not None else "No active subtask",
                        "status": "start"
                    }
                )

            worker_start_time = time.time()

            # Stream action generation start message
            self._send_stream_message(self.task_id, "thinking", "Generating execution actions...")

            # get the next action from the worker
            # Handle case where current_subtask might be None
            subtask_name = self.current_subtask.name if self.current_subtask is not None else "No active subtask"
            subtask_info = self.current_subtask.info if self.current_subtask is not None else ""

            executor_info = self.worker.generate_next_action(
                Tu=instruction,
                search_query=self.search_query,
                subtask=subtask_name,
                subtask_info=subtask_info,
                future_tasks=self.global_state.get_remaining_subtasks(),
                done_task=self.global_state.get_completed_subtasks(),
                obs=self.global_state.get_obs_for_manager(),
            )

            worker_execution_time = time.time() - worker_start_time

            self.global_state.log_operation(
                module="agent",
                operation="worker_execution",
                data={
                    "duration": worker_execution_time,
                    "subtask": self.current_subtask.name if self.current_subtask is not None else "No active subtask" # type: ignore
                }
            )

            # Stream action plan message
            if self.task_id and "executor_plan" in executor_info:
                plan_preview = executor_info["executor_plan"][:100] + "..." if len(executor_info["executor_plan"]) > 100 else executor_info["executor_plan"]
                self._send_stream_message(self.task_id, "action_plan", f"Generate an execution plan: {plan_preview}")

            try:
                grounding_start_time = time.time()
                current_width, current_height = self.global_state.get_screen_size()
                self.grounding.reset_screen_size(current_width, current_height)
                self.grounding.assign_coordinates(executor_info["executor_plan"], observation)
                raw_grounded_action = executor_info["executor_plan"].split("Grounded Action")[-1]
                plan_code = parse_single_code_from_string(raw_grounded_action)
                plan_code = sanitize_code(plan_code)
                plan_code = extract_first_agent_function(plan_code)
                agent: Grounding = self.grounding # type: ignore
                exec_code = eval(plan_code) # type: ignore
                grounding_execution_time = time.time() - grounding_start_time
                
                # 记录grounding执行时间
                self.global_state.log_operation(
                    module="agent",
                    operation="grounding_execution",
                    data={
                        "duration": grounding_execution_time,
                        "content": plan_code
                    }
                )
            except Exception as e:
                if self.global_state.is_cancelled():
                    logger.info("Cancelled during grounding; stopping without action")
                    return {
                        "subtask": "cancelled",
                        "subtask_info": "",
                        "subtask_status": "cancelled",
                        "reflection": "Task was cancelled",
                        "executor_plan": "agent.done()"
                    }, [{"type": "DONE"}]
                logger.error("Error in parsing plan code: %s", e)
                plan_code = "agent.wait(1.0)"
                agent: Grounding = self.grounding # this agent will be used in next code
                exec_code = eval(plan_code) # type: ignore
                
                # 记录grounding错误
                self.global_state.log_operation(
                    module="agent",
                    operation="grounding_error",
                    data={
                        "content": str(e),
                        "fallback_action": plan_code,
                        "raw_grounded_action": (raw_grounded_action if 'raw_grounded_action' in locals() else None),
                        "plan": executor_info.get("executor_plan", "")
                    }
                )

            actions = [exec_code]

            # Stream action execution message
            if actions:
                action_type = actions[0].get("type", "unknown")
                self._send_stream_message(self.task_id, "action", f"Execute an action: {action_type}")

            self.step_count += 1

            # set the should_send_action flag to True if the executor returns an action
            self.should_send_action = True

            # replan on failure
            if "fail" in actions[0]["type"].lower():
                self.requires_replan = True
                self.needs_next_subtask = True

                # assign the failed subtask
                if self.current_subtask is not None:
                    self.global_state.add_failed_subtask(self.current_subtask) # type: ignore
                self.failure_subtask = self.global_state.get_latest_failed_subtask()

                # Stream failure message
                if self.current_subtask is not None:
                    self._send_stream_message(self.task_id, "error", f"Subtask execution failed: {self.current_subtask.name}, will re-plan")
                else:
                    self._send_stream_message(self.task_id, "error", "Subtask execution failed and will be re-planned")

                # 记录失败的子任务
                self.global_state.log_operation(
                    module="agent",
                    operation="subtask_failed",
                    data={
                        "content": str(self.current_subtask) if self.current_subtask is not None else "Unknown subtask",
                        "status": "failed"
                    }
                )

                # reset the step count, executor, and evaluator
                self.reset_executor_state()

                # if more subtasks are remaining, we don't want to send DONE to the environment but move on to the next subtask
                if self.subtasks:
                    self.should_send_action = False

            # replan on subtask completion
            elif "done" in actions[0]["type"].lower():
                self.requires_replan = True
                self.needs_next_subtask = True
                self.failure_subtask = None
                # add completed subtask only if it exists
                if self.current_subtask is not None:
                    self.global_state.add_completed_subtask(self.current_subtask) # type: ignore

                # Stream subtask completion message
                if self.current_subtask is not None:
                    self._send_stream_message(self.task_id, "subtask_complete", f"✅ Subtask completed: {self.current_subtask.name}")
                else:
                    self._send_stream_message(self.task_id, "subtask_complete", "✅ Subtask completed")

                # 记录完成的子任务
                self.global_state.log_operation(
                    module="agent",
                    operation="subtask_completed",
                    data={
                        "content": str(self.current_subtask) if self.current_subtask is not None else "Unknown subtask",
                        "status": "completed"
                    }
                )

                # reset the step count, executor, and evaluator
                self.reset_executor_state()

                # if more subtasks are remaining, we don't want to send DONE to the environment but move on to the next subtask
                if self.subtasks:
                    self.should_send_action = False
                self.subtask_status = "Done"

            self.turn_count += 1

        # reset the should_send_action flag for next iteration
        self.should_send_action = False

        # concatenate the three info dictionaries
        info = {
            **{
                k: v
                for d in [planner_info or {}, executor_info or {}, evaluator_info or {}]
                for k, v in d.items()
            }
        }
        # Handle case where current_subtask might be None
        if self.current_subtask is not None:
            info.update(
                {
                    "subtask": self.current_subtask.name, # type: ignore
                    "subtask_info": self.current_subtask.info, # type: ignore
                    "subtask_status": self.subtask_status,
                }
            )
        else:
            # Handle None case - provide default values
            info.update(
                {
                    "subtask": "No active subtask",
                    "subtask_info": "",
                    "subtask_status": "no_subtask",
                }
            )
        
        # 记录predict函数总执行时间
        predict_total_time = time.time() - predict_start_time
        self.global_state.log_operation(
            module="agent",
            operation="predict_execution",
            data={
                "duration": predict_total_time,
                "step_count": self.step_count,
                "turn_count": self.turn_count,
                "subtask_status": self.subtask_status
            }
        )

        return info, actions # type: ignore

    def update_narrative_memory(self, trajectory: str) -> None:
        """Update narrative memory from task trajectory

        Args:
            trajectory: String containing task execution trajectory
        """
        try:
            reflection_path = os.path.join(
                self.local_kb_path, self.platform, "narrative_memory.json"
            )
            try:
                reflections = json.load(open(reflection_path))
            except:
                reflections = {}

            if self.search_query not in reflections:
                reflection = self.manager.summarize_narrative(trajectory)
                reflections[self.search_query] = reflection

            with open(reflection_path, "w") as f:
                json.dump(reflections, f, indent=2)

        except Exception as e:
            logger.error(f"Failed to update narrative memory: {e}")

    def update_episodic_memory(self, meta_data: Dict, subtask_trajectory: str) -> str:
        """Update episodic memory from subtask trajectory

        Args:
            meta_data: Metadata about current subtask execution
            subtask_trajectory: String containing subtask execution trajectory

        Returns:
            Updated subtask trajectory
        """
        subtask = meta_data["subtask"]
        subtask_info = meta_data["subtask_info"]
        subtask_status = meta_data["subtask_status"]
        # Handle subtask trajectory
        if subtask_status == "Start" or subtask_status == "Done":
            # If it's a new subtask start, finalize the previous subtask trajectory if it exists
            if subtask_trajectory:
                subtask_trajectory += "\nSubtask Completed.\n"
                subtask_key = subtask_trajectory.split(
                    "\n----------------------\n\nPlan:\n"
                )[0]
                try:
                    subtask_path = os.path.join(
                        self.local_kb_path, self.platform, "episodic_memory.json"
                    )
                    kb = json.load(open(subtask_path))
                except:
                    kb = {}
                if subtask_key not in kb.keys():
                    subtask_summarization = self.manager.summarize_episode(
                        subtask_trajectory
                    )
                    kb[subtask_key] = subtask_summarization
                else:
                    subtask_summarization = kb[subtask_key]
                logger.info("subtask_key: %s", subtask_key)
                logger.info("subtask_summarization: %s", subtask_summarization)
                with open(subtask_path, "w") as fout:
                    json.dump(kb, fout, indent=2)
                # Reset for the next subtask
                subtask_trajectory = ""
            # Start a new subtask trajectory
            subtask_trajectory = (
                "Task:\n"
                + self.search_query
                + "\n\nSubtask: "
                + subtask
                + "\nSubtask Instruction: "
                + subtask_info
                + "\n----------------------\n\nPlan:\n"
                + meta_data["executor_plan"]
                + "\n"
            )
        elif subtask_status == "In":
            # Continue appending to the current subtask trajectory if it's still ongoing
            subtask_trajectory += (
                "\n----------------------\n\nPlan:\n"
                + meta_data["executor_plan"]
                + "\n"
            )

        return subtask_trajectory

class AgentSFast(UIAgent):
    """Fast version of AgentS2 that generates a description-based plan with reflection, then grounds to precise coordinates before execution"""

    def __init__(
        self,
        platform: str = platform.system().lower(),
        screen_size: List[int] = [1920, 1080],
        memory_root_path: str = os.getcwd(),
        memory_folder_name: str = "kb_s2",
        kb_release_tag: str = "v0.2.2",
        enable_takeover: bool = False,
        enable_search: bool = True,
        enable_reflection: bool = False,
        tools_config: dict | None = None,
        # enable_reflection: bool = False,
    ):
        """
        Create and initialize an AgentSFast instance, configuring tools, memory paths, and optional features.
        
        Parameters:
            platform (str): Operating system platform identifier (e.g., "darwin", "linux", "windows"); used to scope platform-specific knowledge base.
            screen_size (List[int]): Screen width and height used for grounding calculations.
            memory_root_path (str): Root directory for agent memory storage.
            memory_folder_name (str): Subfolder name under memory_root_path for this agent's knowledge base.
            kb_release_tag (str): Knowledge base release tag used for bookkeeping or compatibility.
            enable_takeover (bool): If True, enable user takeover capabilities in the fast action generator.
            enable_search (bool): If True, enable web/search-related features when registering tools.
            enable_reflection (bool): If True, enable trajectory reflection and a reflection agent to summarize agent behavior.
            tools_config (dict | None): Optional pre-loaded tools configuration; if omitted, configuration is loaded from disk.
        
        """
        platform = platform.lower()
        super().__init__(
            platform,
        )

        self.memory_root_path = memory_root_path
        self.memory_folder_name = memory_folder_name
        self.kb_release_tag = kb_release_tag
        self.screen_size = screen_size
        self.enable_takeover = enable_takeover
        self.enable_search = enable_search
        self.enable_reflection = enable_reflection
        self.task_id = None  # Will be set when task starts

        if tools_config is not None:
            self.tools_config = tools_config
            # Create the dictionary mapping from the list-based config
            self.Tools_dict = {}
            for tool in self.tools_config["tools"]:
                tool_name = tool["tool_name"]
                # Create a copy of the tool's config to avoid modifying the original
                config_copy = tool.copy()
                # Rename 'model_name' to 'model' for consistency in downstream use
                if 'model_name' in config_copy:
                    config_copy['model'] = config_copy.pop('model_name')
                # Remove tool_name as it's now the key
                config_copy.pop('tool_name', None)
                self.Tools_dict[tool_name] = config_copy
        else:
            self.tools_config, self.Tools_dict = load_config()

        # Initialize agent's knowledge base path
        self.local_kb_path = os.path.join(
            self.memory_root_path, self.memory_folder_name
        )

        # Check if knowledge base exists
        kb_platform_path = os.path.join(self.local_kb_path, self.platform)
        if not os.path.exists(kb_platform_path):
            print(f"Warning: Knowledge base for {self.platform} platform not found in {self.local_kb_path}")
            os.makedirs(kb_platform_path, exist_ok=True)
            print(f"Created directory: {kb_platform_path}")
        else:
            print(f"Found local knowledge base path: {kb_platform_path}")

    def reset(self) -> None:
        """
        Reinitialize the fast-agent components and reset internal runtime state.
        
        Initializes and registers the fast action generator tool (and traj_reflector if reflection is enabled), configures search/auth parameters from tool configuration, creates or updates the grounding subsystem with resolved grounding dimensions, resets counters and runtime references (step_count, turn_count, latest_action, global_state), and propagates the current task_id to any registered tools.
        """
        # Initialize the fast action generator tool
        self.fast_action_generator = Tools()
        self.fast_action_generator_tool = "fast_action_generator_with_takeover" if self.enable_takeover else "fast_action_generator"

        # Get tool configuration from tools_config
        tool_config = None
        for tool in self.tools_config["tools"]:
            if tool["tool_name"] == self.fast_action_generator_tool:
                tool_config = tool
                break

        # Prepare tool parameters
        tool_params = {}

        # First check global search switch
        if not self.enable_search:
            # If global search is disabled, force disable search for this tool
            tool_params["enable_search"] = False
            logger.info(f"Configuring {self.fast_action_generator_tool} with search DISABLED (global switch off)")
        else:
            # If global search is enabled, check tool-specific config
            if tool_config and "enable_search" in tool_config:
                # Use enable_search from config file
                enable_search = tool_config.get("enable_search", False)
                tool_params["enable_search"] = enable_search
                tool_params["search_provider"] = tool_config.get("search_provider", "bocha")
                tool_params["search_model"] = tool_config.get("search_model", "")

                logger.info(f"Configuring {self.fast_action_generator_tool} with search enabled: {enable_search} (from config)")

        # Get base config from Tools_dict
        tool_config = self.Tools_dict[self.fast_action_generator_tool].copy()
        provider = tool_config.get("provider")
        model = tool_config.get("model")

        # Merge with search-related parameters
        all_params = {**tool_config, **tool_params}

        # Remove provider and model from all_params to avoid duplicate arguments
        all_params.pop("provider", None)
        all_params.pop("model", None)

        auth_keys = ['api_key', 'base_url', 'endpoint_url', 'azure_endpoint', 'api_version']
        for key in auth_keys:
            if key in all_params:
                logger.info(f"AgentSFast.reset: Setting {key} for fast_action_generator_tool")

        # Register the tool with all parameters
        self.fast_action_generator.register_tool(
            self.fast_action_generator_tool,
            provider,
            model,
            **all_params
        )

        if self.enable_reflection:
            self.reflection_agent = Tools()

            # Get base config from Tools_dict
            reflector_tool_config = self.Tools_dict["traj_reflector"].copy()
            reflector_provider = reflector_tool_config.get("provider")
            reflector_model = reflector_tool_config.get("model")

            # Remove provider and model from reflector_tool_config to avoid duplicate arguments
            reflector_tool_config.pop("provider", None)
            reflector_tool_config.pop("model", None)

            auth_keys = ['api_key', 'base_url', 'endpoint_url', 'azure_endpoint', 'api_version']
            for key in auth_keys:
                if key in reflector_tool_config:
                    logger.info(f"AgentSFast.reset: Setting {key} for traj_reflector")

            # Register the reflection tool
            self.reflection_agent.register_tool(
                "traj_reflector", self.Tools_dict["traj_reflector"]["provider"],
                self.Tools_dict["traj_reflector"]["model"],
                **reflector_tool_config
            )
            self.reflections = []
            self.planner_history = []

        # Use normal Grounding (description -> coordinates) instead of direct coordinate execution
        self.grounding = Grounding(
            Tools_dict=self.Tools_dict,
            platform=self.platform,
            width=self.screen_size[0],
            height=self.screen_size[1]
        )

        # Reset state variables
        self.step_count: int = 0
        self.turn_count: int = 0
        # Use task-specific registry if task_id is available, otherwise fall back to global registry
        if self.task_id:
            self.global_state: GlobalState = Registry.get_from_context("GlobalStateStore", self.task_id) # type: ignore
        else:
            self.global_state: GlobalState = Registry.get("GlobalStateStore") # type: ignore
        self.latest_action = None
        self.last_exec_plan_code: Optional[str] = None
        self.last_exec_repeat: int = 0
        self.raw_grounded_action: Optional[str] = None

        # Pass task_id to tools and components if available
        self.fast_action_generator.task_id = self.task_id
        if self.enable_reflection and hasattr(self, 'reflection_agent'):
            self.reflection_agent.task_id = self.task_id
        # Set task_id for grounding component
        if hasattr(self, 'grounding') and hasattr(self.grounding, 'set_task_id'):
            self.grounding.set_task_id(self.task_id)

    def set_task_id(self, task_id: str) -> None:
        """
        Store the task identifier on the agent and propagate it to subcomponents that use it.

        Parameters:
            task_id (str): Identifier for the active task; assigned to this agent and, if present, to
                `fast_action_generator` and `reflection_agent`.
        """
        self.task_id = task_id
        # Also set task_id for components if they exist
        if hasattr(self, 'fast_action_generator') and self.fast_action_generator:
            self.fast_action_generator.task_id = task_id
        if hasattr(self, 'reflection_agent') and self.reflection_agent:
            self.reflection_agent.task_id = task_id
        # Set task_id for grounding component
        if hasattr(self, 'grounding') and hasattr(self.grounding, 'set_task_id'):
            self.grounding.set_task_id(task_id)

    def predict(self, instruction: str, observation: Dict) -> Tuple[Dict, List[str]]:
        """
        Generate the next executor plan and corresponding actions using the configured fast action generator.

        Parameters:
        	instruction (str): Natural language task description.
        	observation (Dict): Current UI state; must include a "screenshot" entry with the screen image.

        Returns:
        	executor_info (dict): Contains at least the keys `executor_plan` (raw plan text), `reflection` (reflection text or empty string), and `plan_code` (the latest extracted/used action code).
        	actions (List[dict]): List of action dictionaries produced by grounding execution; typically a single action dict describing the operation to perform.
        """
        # Check for cancellation before starting prediction
        if self.global_state.is_cancelled():
            logger.info("AgentSFast prediction cancelled by user request")
            return {
                "executor_plan": "agent.done()",
                "reflection": "Task was cancelled",
                "plan_code": "agent.done()"
            }, [{"type": "DONE"}]

        import time
        predict_start_time = time.time()

        fast_action_start_time = time.time()

        reflection = None
        if self.enable_reflection:
            if self.turn_count == 0:
                text_content = textwrap.dedent(f"""
                    Task Description: {instruction}
                    """)
                self.reflection_agent.tools["traj_reflector"].llm_agent.add_message(
                    text_content + "\n\nThe initial screen is provided. No action has been taken yet.",
                    image_content=observation["screenshot"],
                    role="user")
                self.global_state.add_agent_log({
                    "type": "passive",
                    "content": "Reflection: " + text_content + "\n\nThe initial screen is provided. No action has been taken yet."
                })
            else:
                agent_log = agent_log_to_string(self.global_state.get_agent_log())
                text_content = f"Please refer to the agent log to understand the progress and context of the task so far.\n{agent_log}"

                reflection_start = time.time()
                reflection, total_tokens, cost_string = self.reflection_agent.execute_tool(
                    "traj_reflector", {
                        "str_input": text_content,
                        "img_input": observation["screenshot"]
                    })
                reflection = str(reflection)
                self.reflection_agent.reset("traj_reflector")
                self.global_state.add_agent_log({
                    "type": "passive",
                    "content": "Reflection: " + reflection
                })
                logger.info(f"Trajectory reflector tokens: {total_tokens}, cost: {cost_string}")
                reflection_time = time.time() - reflection_start
                logger.info(f"[Timing] AgentSFast.traj_reflector execution time: {reflection_time:.2f} seconds")
                self.reflections.append(reflection)
                logger.info("REFLECTION: %s", reflection)
                self.global_state.log_operation(
                    module="agent",
                    operation="reflection",
                    data={
                        "tokens": total_tokens,
                        "cost": cost_string,
                        "content": reflection,
                        "duration": reflection_time
                    })

        agent_log = agent_log_to_string(self.global_state.get_agent_log())
        generator_message = textwrap.dedent(f"""
            Task Description: {instruction}
        """)

        generator_message += f"\n\nPlease refer to the agent log to understand the progress and context of the task so far.\n{agent_log}"

        fast_action_start_time = time.time()

        # Stream action generation start message
        self._send_stream_message(self.task_id, "thinking", "Generating execution actions quickly...")

        plan, total_tokens, cost_string = self.fast_action_generator.execute_tool(
            self.fast_action_generator_tool,
            {
                "str_input": generator_message,
                "img_input": observation["screenshot"]
            }
        )
        self.fast_action_generator.reset(self.fast_action_generator_tool)

        fast_action_execution_time = time.time() - fast_action_start_time

        self.global_state.log_operation(
            module="agent",
            operation="fast_planning_execution",
            data={
                "duration": fast_action_execution_time,
                "tokens": total_tokens,
                "cost": cost_string
            }
        )

        # Stream action plan message
        if self.task_id:
            plan_preview = plan[:100] + "..." if len(plan) > 100 else plan
            self._send_stream_message(self.task_id, "action_plan", f"Quickly generate execution plans: {plan_preview}")

        logger.info("Fast Action Plan: %s", plan)

        current_width, current_height = self.global_state.get_screen_size()
        self.grounding.reset_screen_size(current_width, current_height)
        try:
            grounding_start_time = time.time()
            self.raw_grounded_action = plan.split("Grounded Action")[-1]
            plan_code = parse_single_code_from_string(self.raw_grounded_action)
            self.grounding.assign_coordinates(plan, observation)
            plan_code = sanitize_code(plan_code)
            plan_code = extract_first_agent_function(plan_code)
            agent: Grounding = self.grounding  # type: ignore
            exec_code = eval(plan_code)  # type: ignore
            grounding_execution_time = time.time() - grounding_start_time

            self.global_state.log_operation(
                module="agent",
                operation="fast_grounding_execution",
                data={
                    "duration": grounding_execution_time,
                    "content": plan_code
                }
            )

            actions = [exec_code]
            self.latest_action = plan_code

            if plan_code == (self.last_exec_plan_code or None):
                self.last_exec_repeat += 1
            else:
                self.last_exec_plan_code = plan_code
                self.last_exec_repeat = 1
            if self.last_exec_repeat >= 3:
                warning_msg = f"Action repeated {self.last_exec_repeat} times, possible stuck: {plan_code}"
                logger.warning(warning_msg)
                self.global_state.add_agent_log({
                    "type": "warning",
                    "content": warning_msg
                })
        except Exception as e:
            logger.error("Error in parsing action code: %s", e)
            self.global_state.add_agent_log({
                "type": "Error in parsing action code",
                "content": f"error={str(e)}; latest_grounded_action={self.raw_grounded_action}"
            })
            agent: Grounding = self.grounding  # type: ignore
            exec_code = eval("agent.wait(1000)")  # type: ignore
            actions = [exec_code]
            self.latest_action = "agent.wait(1000)"
            
            if self.latest_action == (self.last_exec_plan_code or None):
                self.last_exec_repeat += 1
            else:
                self.last_exec_plan_code = self.latest_action
                self.last_exec_repeat = 1
            if self.last_exec_repeat >= 3:
                warning_msg = f"Action repeated {self.last_exec_repeat} times, possible stuck: {self.raw_grounded_action}"
                logger.warning(warning_msg)
                self.global_state.add_agent_log({
                    "type": "warning",
                    "content": warning_msg
                })

            self.global_state.log_operation(
                module="agent",
                operation="fast_action_error",
                data={
                    "content": str(e),
                    "fallback_action": "agent.wait(1000)",
                    "raw_grounded_action": self.raw_grounded_action,
                    "plan": plan
                }
            )

        self.step_count += 1
        self.turn_count += 1

        # Stream action execution message
        if actions:
            action_type = actions[0].get("type", "unknown")
            self._send_stream_message(self.task_id, "action", f"Execute an action: {action_type}")

        executor_info = {
            "executor_plan": plan,
            "reflection": "",
            "plan_code": self.latest_action
        }

        predict_total_time = time.time() - predict_start_time
        self.global_state.log_operation(
            module="agent",
            operation="predict_execution_fast_direct",
            data={
                "duration": predict_total_time,
                "step_count": self.step_count,
                "turn_count": self.turn_count
            }
        )

        return executor_info, actions


import logging
import re
import textwrap
from typing import Dict, List
import platform
import os
import json

from gui_agents.core.knowledge import KnowledgeBase
from gui_agents.utils.common_utils import (
    Node,
    extract_first_agent_function,
    parse_single_code_from_string,
    sanitize_code,
    agent_log_to_string,
)
from gui_agents.tools.tools import Tools
from gui_agents.store.registry import Registry
from gui_agents.agents.global_state import GlobalState

logger = logging.getLogger("desktopenv.agent")


class Worker:

    def __init__(
        self,
        Tools_dict: Dict,
        local_kb_path: str,
        platform: str = platform.system().lower(),
        enable_reflection: bool = True,
        use_subtask_experience: bool = True,
        enable_takeover: bool = False,
        enable_search: bool = True,
        tools_config: Dict = {},
    ):
        """
        Initialize a Worker that generates executor actions using the provided tools, local knowledge base, and optional reflection, episodic experience, takeover, and search features.
        
        Parameters:
            Tools_dict (Dict): Mapping of tool names to tool instances/configurations used by the Worker.
            local_kb_path (str): Filesystem path to the local knowledge base to use for retrieval.
            platform (str): Operating system identifier the agent runs on (e.g., 'darwin', 'linux', 'windows').
            enable_reflection (bool): If True, enable trajectory reflection generation and use its output when producing actions.
            use_subtask_experience (bool): If True, attempt to retrieve and incorporate episodic/subtask experience on the first turn.
            enable_takeover (bool): If True, use the takeover-capable action generator tool when producing actions.
            enable_search (bool): Global switch that forces search-enabled tools to run with search disabled when False.
            tools_config (Dict): Tools configuration mapping; if None, the Worker loads tools_config.json from the package tools directory.
        """
        # super().__init__(engine_params, platform)
        self.platform = platform

        self.local_kb_path = local_kb_path
        self.Tools_dict = Tools_dict
        self.enable_takeover = enable_takeover
        self.enable_search = enable_search  # Store global search switch

        # If tools_config is not provided, load it from file
        if tools_config is None:
            tools_config_path = os.path.join(
                os.path.dirname(os.path.dirname(__file__)), "tools",
                "tools_config.json")
            with open(tools_config_path, "r") as f:
                self.tools_config = json.load(f)
        else:
            self.tools_config = tools_config

        self.enable_reflection = enable_reflection
        self.use_subtask_experience = use_subtask_experience
        # GlobalState will be initialized in reset() method when task_id is available
        self.global_state: GlobalState = None  # type: ignore
        self.reset()

    def reset(self):

        """
        Initialize the worker's tool agents, knowledge base, and internal state for a new task session.
        
        This method registers the action generator (with optional takeover variant), trajectory reflector, and embedding engine using a local helper that merges tool configuration with any overrides and propagates authentication parameters; it initializes the KnowledgeBase with the embedding engine and toolkit, configures search-related parameters for the action generator according to global and per-tool settings, and resets runtime state fields (turn count, histories, reflections, cost tracking, screenshot inputs, planner history, latest action, trajectory length limit, and task_id).
        """
        def _register(tools_instance, tool_name, **override_kwargs):
            config = self.Tools_dict.get(tool_name, {}).copy()
            provider = config.pop("provider", None)
            model = config.pop("model", None)

            # Merge with any explicit overrides
            config.update(override_kwargs)

            auth_params = {}
            auth_keys = ['api_key', 'base_url', 'endpoint_url', 'azure_endpoint', 'api_version']
            for key in auth_keys:
                if key in config:
                    auth_params[key] = config[key]
                    logger.info(f"Worker._register: Setting {key} for tool '{tool_name}'")

            all_params = {**config, **auth_params}

            logger.info(f"Worker._register: Registering tool '{tool_name}' with provider '{provider}', model '{model}'")
            tools_instance.register_tool(tool_name, provider, model, **all_params)

        self.generator_agent = Tools()
        self.action_generator_tool = "action_generator_with_takeover" if self.enable_takeover else "action_generator"

        # Get tool configuration from tools_config
        tool_config = None
        for tool in self.tools_config["tools"]:
            if tool["tool_name"] == self.action_generator_tool:
                tool_config = tool
                break

        # Prepare tool parameters
        tool_params = {}

        # First check global search switch
        if not self.enable_search:
            # If global search is disabled, force disable search for this tool
            tool_params["enable_search"] = False
            logger.info(
                f"Configuring {self.action_generator_tool} with search DISABLED (global switch off)"
            )
        else:
            # If global search is enabled, check tool-specific config
            if tool_config and "enable_search" in tool_config:
                # Use enable_search from config file
                enable_search = tool_config.get("enable_search", False)
                tool_params["enable_search"] = enable_search
                tool_params["search_provider"] = tool_config.get(
                    "search_provider", "bocha")
                tool_params["search_model"] = tool_config.get(
                    "search_model", "")

                logger.info(
                    f"Configuring {self.action_generator_tool} with search enabled: {enable_search} (from config)"
                )

        # Register the tool with parameters
        _register(self.generator_agent, self.action_generator_tool, **tool_params)

        self.reflection_agent = Tools()
        _register(self.reflection_agent, "traj_reflector")

        self.embedding_engine = Tools()
        _register(self.embedding_engine, "embedding")

        self.knowledge_base = KnowledgeBase(
            embedding_engine=self.embedding_engine,
            Tools_dict=self.Tools_dict,
            local_kb_path=self.local_kb_path,
            platform=self.platform,
        )

        self.turn_count = 0
        self.worker_history = []
        self.reflections = []
        self.cost_this_turn = 0
        self.screenshot_inputs = []
        self.planner_history = []
        self.latest_action = None
        self.max_trajector_length = 8
        self.task_id = None  # Will be set by agent

    def set_task_id(self, task_id: str) -> None:
        """Set the task identifier and update global state reference"""
        self.task_id = task_id
        # Update global state reference with task-specific registry
        self.global_state = Registry.get_from_context("GlobalStateStore", task_id)  # type: ignore

    def generate_next_action(
        self,
        Tu: str,
        search_query: str,
        subtask: str,
        subtask_info: str,
        future_tasks: List[Node],
        done_task: List[Node],
        obs: Dict,
        running_state: str = "running",
    ) -> Dict:
        """
        Generate the next executor action plan and related metadata for the current subtask given the observation and context.

        Parameters:
            Tu (str): Full task description or task context.
            search_query (str): Search string used for retrieving episodic/subtask experience.
            subtask (str): Current subtask instruction/description to complete.
            subtask_info (str): Additional information or constraints for the current subtask.
            future_tasks (List[Node]): List of upcoming task nodes (used for context in planning).
            done_task (List[Node]): List of completed task nodes.
            obs (Dict): Current observation dictionary; must include a "screenshot" key with the current screen image.
            running_state (str): Current executor running state (default "running").

        Returns:
            Dict: Executor information containing:
                - "current_subtask" (str): The provided subtask.
                - "current_subtask_info" (str): The provided subtask_info.
                - "executor_plan" (str): The raw plan produced by the action generator.
                - "reflection" (str|None): Reflection text produced by the trajectory reflector, or None if reflection is disabled.
        """
        # Check for cancellation before starting action generation
        if self.global_state.is_cancelled():
            logger.info("Worker action generation cancelled by user request")
            return {
                "current_subtask": subtask,
                "current_subtask_info": subtask_info,
                "executor_plan": "agent.done()",
                "reflection": "Task was cancelled"
            }

        import time
        action_start = time.time()

        # Log the result of the previous hardware action, which is the current observation.
        if self.turn_count > 0 and self.latest_action:
            self.global_state.add_agent_log({
                "type":
                    "passive",
                "content":
                    f"Hardware action `{self.latest_action}` has been executed. The result is reflected in the current screenshot."
            })

        # Get RAG knowledge, only update system message at t=0
        if self.turn_count == 0:
            if self.use_subtask_experience:
                subtask_query_key = ("Task:\n" + search_query +
                                     "\n\nSubtask: " + subtask +
                                     "\nSubtask Instruction: " + subtask_info)
                retrieve_start = time.time()
                retrieved_similar_subtask, retrieved_subtask_experience, total_tokens, cost_string = (
                    self.knowledge_base.retrieve_episodic_experience(
                        subtask_query_key))
                logger.info(
                    f"Retrieve episodic experience tokens: {total_tokens}, cost: {cost_string}"
                )
                retrieve_time = time.time() - retrieve_start
                logger.info(
                    f"[Timing] Worker.retrieve_episodic_experience execution time: {retrieve_time:.2f} seconds"
                )

                # Dirty fix to replace id with element description during subtask retrieval
                pattern = r"\(\d+"
                retrieved_subtask_experience = re.sub(
                    pattern, "(element_description",
                    retrieved_subtask_experience)
                retrieved_subtask_experience = retrieved_subtask_experience.replace(
                    "_id", "_description")

                logger.info(
                    "SIMILAR SUBTASK EXPERIENCE: %s",
                    retrieved_similar_subtask + "\n" +
                    retrieved_subtask_experience.strip(),
                )
                self.global_state.log_operation(
                    module="worker",
                    operation="Worker.retrieve_episodic_experience",
                    data={
                        "tokens":
                            total_tokens,
                        "cost":
                            cost_string,
                        "content":
                            "Retrieved similar subtask: " +
                            retrieved_similar_subtask + "\n" +
                            "Retrieved subtask experience: " +
                            retrieved_subtask_experience.strip(),
                        "duration":
                            retrieve_time
                    })
                Tu += "\nYou may refer to some similar subtask experience if you think they are useful. {}".format(
                    retrieved_similar_subtask + "\n" +
                    retrieved_subtask_experience)

            prefix_message = f"SUBTASK_DESCRIPTION is {subtask}\n\nTASK_DESCRIPTION is {Tu}\n\nFUTURE_TASKS is {', '.join([f.name for f in future_tasks])}\n\nDONE_TASKS is {', '.join(d.name for d in done_task)}"

        # Reflection generation does not add its own response, it only gets the trajectory
        reflection = None
        if self.enable_reflection:
            # Load the initial subtask info
            if self.turn_count == 0:
                text_content = textwrap.dedent(f"""
                    Subtask Description: {subtask}
                    Subtask Information: {subtask_info}
                    Current Trajectory below:
                    """)
                self.reflection_agent.tools["traj_reflector"].llm_agent.add_message(
                    text_content +
                    "\n\nThe initial screen is provided. No action has been taken yet.",
                    image_content=obs["screenshot"],
                    role="user")

            else:
                if self.planner_history and self.planner_history[-1] is not None:
                    text_content = self.clean_worker_generation_for_reflection(
                        self.planner_history[-1])
                else:
                    text_content = "No previous action available for reflection"

                reflection_start = time.time()
                reflection, total_tokens, cost_string = self.reflection_agent.execute_tool(
                    "traj_reflector", {
                        "str_input": text_content,
                        "img_input": obs["screenshot"]
                    })
                logger.info(
                    f"Trajectory reflector tokens: {total_tokens}, cost: {cost_string}"
                )
                reflection_time = time.time() - reflection_start
                logger.info(
                    f"[Timing] Worker.traj_reflector execution time: {reflection_time:.2f} seconds"
                )
                self.reflections.append(reflection)
                logger.info("REFLECTION: %s", reflection)
                self.global_state.log_operation(module="manager",
                                                operation="reflection",
                                                data={
                                                    "tokens": total_tokens,
                                                    "cost": cost_string,
                                                    "content": reflection,
                                                    "duration": reflection_time
                                                })

        generator_message = ""

        # Only provide subinfo in the very first message to avoid over influence and redundancy
        if self.turn_count == 0:
            generator_message += prefix_message
            generator_message += f"Remember only complete the subtask: {subtask}\n"
            generator_message += f"You can use this extra information for completing the current subtask: {subtask_info}.\n"
        else:
            agent_log = agent_log_to_string(self.global_state.get_agent_log())
            generator_message += f"\nYour previous action was: {self.latest_action}\n"
            generator_message += (
                f"\nYou may use this reflection on the previous action and overall trajectory: {reflection}\n"
                if reflection and self.turn_count > 0 else "")
            generator_message += f"Please refer to the agent log to understand the progress and context of the task so far.\n{agent_log}"

        action_generator_start = time.time()
        plan, total_tokens, cost_string = self.generator_agent.execute_tool(
            "action_generator_with_takeover"
            if self.enable_takeover else "action_generator", {
                "str_input": generator_message,
                "img_input": obs["screenshot"]
            })
        logger.info(
            f"Action generator tokens: {total_tokens}, cost: {cost_string}")
        action_generator_time = time.time() - action_generator_start
        logger.info(
            f"[Timing] Worker.action_generator execution time: {action_generator_time:.2f} seconds"
        )

        self.planner_history.append(plan)
        logger.info("Action Plan: %s", plan)
        self.global_state.log_operation(module="worker",
                                        operation="action_plan",
                                        data={
                                            "tokens": total_tokens,
                                            "cost": cost_string,
                                            "content": plan,
                                            "duration": action_generator_time
                                        })

        # Add the generated plan to the agent log as passive memory
        self.global_state.add_agent_log({"type": "passive", "content": plan})

        try:
            action_code = parse_single_code_from_string(
                plan.split("Grounded Action")[-1])
            action_code = sanitize_code(action_code)
            self.latest_action = extract_first_agent_function(action_code)
        except Exception as e:
            logger.warning(f"Failed to parse action from plan: {e}")
            self.latest_action = None

        executor_info = {
            "current_subtask": subtask,
            "current_subtask_info": subtask_info,
            "executor_plan": plan,
            "reflection": reflection,
        }
        self.turn_count += 1

        self.screenshot_inputs.append(obs["screenshot"])

        return executor_info

    # Removes the previous action verification, and removes any extraneous grounded actions
    def clean_worker_generation_for_reflection(self,
                                               worker_generation: str) -> str:
        # Remove the previous action verification
        res = worker_generation[worker_generation.find("(Screenshot Analysis)"
                                                      ):]
        action = extract_first_agent_function(worker_generation)
        # Cut off extra grounded actions
        res = res[:res.find("(Grounded Action)")]
        res += f"(Grounded Action)\n```python\n{action}\n```\n"
        return res


import os
import json
import backoff
import requests
import numpy as np
from anthropic import Anthropic
from openai import (
    AzureOpenAI,
    APIConnectionError,
    APIError,
    AzureOpenAI,
    OpenAI,
    RateLimitError,
)
from google import genai
from google.genai import types
from zhipuai import ZhipuAI
from groq import Groq
import boto3
from typing import List, Dict, Any, Optional, Union, Tuple

class ModelPricing:
    def __init__(self, pricing_file: str = "model_pricing.json"):
        self.pricing_file = pricing_file
        self.pricing_data = self._load_pricing()
    
    def _load_pricing(self) -> Dict:
        if os.path.exists(self.pricing_file):
            try:
                with open(self.pricing_file, 'r', encoding='utf-8') as f:
                    return json.load(f)
            except Exception as e:
                print(f"Warning: Failed to load pricing file {self.pricing_file}: {e}")
        
        return {
            "default": {"input": 0, "output": 0}
        }
    
    def get_price(self, model: str) -> Dict[str, float]:
        # Handle nested pricing data structure
        if "llm_models" in self.pricing_data:
            # Iterate through all LLM model categories
            for category, models in self.pricing_data["llm_models"].items():
                # Direct model name matching
                if model in models:
                    pricing = models[model]
                    return self._parse_pricing(pricing)
                
                # Fuzzy matching for model names
                for model_name in models:
                    if model_name in model or model in model_name:
                        pricing = models[model_name]
                        return self._parse_pricing(pricing)
        
        # Handle embedding models
        if "embedding_models" in self.pricing_data:
            for category, models in self.pricing_data["embedding_models"].items():
                if model in models:
                    pricing = models[model]
                    return self._parse_pricing(pricing)
                
                for model_name in models:
                    if model_name in model or model in model_name:
                        pricing = models[model_name]
                        return self._parse_pricing(pricing)
        
        # Default pricing
        return {"input": 0, "output": 0}
    
    def _parse_pricing(self, pricing: Dict[str, str]) -> Dict[str, float]:
        """Parse pricing strings and convert to numeric values"""
        result = {}
        
        for key, value in pricing.items():
            if isinstance(value, str):
                # Remove currency symbols and units, convert to float
                clean_value = value.replace('$', '').replace('￥', '').replace(',', '')
                try:
                    result[key] = float(clean_value)
                except ValueError:
                    result[key] = 0.0
            else:
                result[key] = float(value) if value else 0.0
        
        return result
    
    def calculate_cost(self, model: str, input_tokens: int, output_tokens: int) -> float:
        pricing = self.get_price(model)
        input_cost = (input_tokens / 1000000) * pricing["input"]
        output_cost = (output_tokens / 1000000) * pricing["output"]
        return input_cost + output_cost

# Initialize pricing manager with correct pricing file path
pricing_file = os.path.join(os.path.dirname(__file__), 'model_pricing.json')
pricing_manager = ModelPricing(pricing_file)

def extract_token_usage(response, provider: str) -> Tuple[int, int]:
    if "-" in provider:
        api_type, vendor = provider.split("-", 1)
    else:
        api_type, vendor = "llm", provider

    if api_type == "llm":
        if vendor in ["openai", "qwen", "deepseek", "doubao", "siliconflow", "monica", "vllm", "groq", "zhipu", "gemini", "openrouter", "azureopenai", "huggingface", "exa"]:
            if hasattr(response, 'usage') and response.usage:
                return response.usage.prompt_tokens, response.usage.completion_tokens
        
        elif vendor == "anthropic":
            if hasattr(response, 'usage') and response.usage:
                return response.usage.input_tokens, response.usage.output_tokens
        
        elif vendor == "bedrock":
            if isinstance(response, dict) and "usage" in response:
                usage = response["usage"]
                return usage.get("input_tokens", 0), usage.get("output_tokens", 0)
    
    elif api_type == "embedding":
        if vendor in ["openai", "azureopenai", "qwen", "doubao"]:
            if hasattr(response, 'usage') and response.usage:
                return response.usage.prompt_tokens, 0
        
        elif vendor == "jina":
            if isinstance(response, dict) and "usage" in response:
                total_tokens = response["usage"].get("total_tokens", 0)
                return total_tokens, 0
        
        elif vendor == "gemini":
            if hasattr(response, 'usage') and response.usage:
                return response.usage.prompt_tokens, 0

    return 0, 0

def calculate_tokens_and_cost(response, provider: str, model: str) -> Tuple[List[int], float]:
    input_tokens, output_tokens = extract_token_usage(response, provider)
    total_tokens = input_tokens + output_tokens
    cost = pricing_manager.calculate_cost(model, input_tokens, output_tokens)
    
    return [input_tokens, output_tokens, total_tokens], cost

class LMMEngine:
    pass

# ==================== LLM ====================

class LMMEngineOpenAI(LMMEngine):
    def __init__(
        self, base_url=None, api_key=None, model=None, rate_limit=-1, **kwargs
    ):
        assert model is not None, "model must be provided"
        self.model = model
        self.provider = "llm-openai"

        api_key = api_key or os.getenv("OPENAI_API_KEY")
        if api_key is None:
            raise ValueError(
                "An API Key needs to be provided in either the api_key parameter or as an environment variable named OPENAI_API_KEY"
            )

        self.base_url = base_url

        self.api_key = api_key
        self.request_interval = 0 if rate_limit == -1 else 60.0 / rate_limit

        if not self.base_url:
            self.llm_client = OpenAI(api_key=self.api_key)
        else:
            self.llm_client = OpenAI(base_url=self.base_url, api_key=self.api_key)

    @backoff.on_exception(
        backoff.expo, (APIConnectionError, APIError, RateLimitError), max_time=60
    )
    def generate(self, messages, temperature=0.0, max_new_tokens=None, **kwargs):
        """Generate the next message based on previous messages"""
        response = self.llm_client.chat.completions.create(
            model=self.model,
            messages=messages,
            max_tokens=max_new_tokens if max_new_tokens else 4096,
            temperature=temperature,
            **kwargs,
        )
        
        content = response.choices[0].message.content
        total_tokens, cost = calculate_tokens_and_cost(response, self.provider, self.model)
        
        return content, total_tokens, cost


class LMMEngineQwen(LMMEngine):
    def __init__(
        self, base_url=None, api_key=None, model=None, rate_limit=-1, enable_thinking=False, **kwargs
    ):
        assert model is not None, "model must be provided"
        self.model = model
        self.enable_thinking = enable_thinking
        self.provider = "llm-qwen"

        api_key = api_key or os.getenv("DASHSCOPE_API_KEY")
        if api_key is None:
            raise ValueError(
                "An API Key needs to be provided in either the api_key parameter or as an environment variable named DASHSCOPE_API_KEY"
            )

        self.base_url = base_url or "https://dashscope.aliyuncs.com/compatible-mode/v1"
        self.api_key = api_key
        self.request_interval = 0 if rate_limit == -1 else 60.0 / rate_limit

        self.llm_client = OpenAI(base_url=self.base_url, api_key=self.api_key)

    @backoff.on_exception(
        backoff.expo, (APIConnectionError, APIError, RateLimitError), max_time=60
    )
    def generate(self, messages, temperature=0.0, max_new_tokens=None, **kwargs):
        """Generate the next message based on previous messages"""
        # For Qwen3 models, we need to handle thinking mode
        extra_body = {}
        if self.model.startswith("qwen3") and not self.enable_thinking:
            extra_body["enable_thinking"] = False

        response = self.llm_client.chat.completions.create(
            model=self.model,
            messages=messages,
            max_tokens=max_new_tokens if max_new_tokens else 4096,
            temperature=temperature,
            **extra_body,
            **kwargs,
        )
        
        content = response.choices[0].message.content
        total_tokens, cost = calculate_tokens_and_cost(response, self.provider, self.model)
        
        return content, total_tokens, cost


class LMMEngineDoubao(LMMEngine):
    def __init__(
        self, base_url=None, api_key=None, model=None, rate_limit=-1, **kwargs
    ):
        assert model is not None, "model must be provided"
        self.model = model
        self.provider = "llm-doubao"

        api_key = api_key or os.getenv("ARK_API_KEY")
        if api_key is None:
            raise ValueError(
                "An API Key needs to be provided in either the api_key parameter or as an environment variable named ARK_API_KEY"
            )

        self.base_url = base_url or "https://ark.cn-beijing.volces.com/api/v3"
        self.api_key = api_key
        self.request_interval = 0 if rate_limit == -1 else 60.0 / rate_limit

        self.llm_client = OpenAI(base_url=self.base_url, api_key=self.api_key)

    @backoff.on_exception(
        backoff.expo, (APIConnectionError, APIError, RateLimitError), max_time=60
    )
    def generate(self, messages, temperature=0.0, max_new_tokens=None, **kwargs):
        """Generate the next message based on previous messages"""
        try:
            response = self.llm_client.chat.completions.create(
                model=self.model,
                messages=messages,
                max_tokens=max_new_tokens if max_new_tokens else 4096,
                temperature=temperature,
                extra_body={
                    "thinking": {
                        "type": "disabled",
                        # "type": "enabled",
                        # "type": "auto",
                    }
                },
                **kwargs,
            )
        except APIError as e:
            message = str(e)
            if "Total tokens of image and text exceed max message tokens" in message or (
                "max message tokens" in message and "Total tokens" in message
            ):
                raise ValueError(f"Doubao token limit exceeded: {message}") from e
            raise
        
        content = response.choices[0].message.content
        total_tokens, cost = calculate_tokens_and_cost(response, self.provider, self.model)
        
        return content, total_tokens, cost


class LMMEngineAnthropic(LMMEngine):
    def __init__(
        self, base_url=None, api_key=None, model=None, thinking=False, **kwargs
    ):
        assert model is not None, "model must be provided"
        self.model = model
        self.thinking = thinking
        self.provider = "llm-anthropic"

        api_key = api_key or os.getenv("ANTHROPIC_API_KEY")
        if api_key is None:
            raise ValueError(
                "An API Key needs to be provided in either the api_key parameter or as an environment variable named ANTHROPIC_API_KEY"
            )

        self.api_key = api_key

        self.llm_client = Anthropic(api_key=self.api_key)

    @backoff.on_exception(
        backoff.expo, (APIConnectionError, APIError, RateLimitError), max_time=60
    )
    def generate(self, messages, temperature=0.0, max_new_tokens=None, **kwargs):
        """Generate the next message based on previous messages"""
        if self.thinking:
            response = self.llm_client.messages.create(
                system=messages[0]["content"][0]["text"],
                model=self.model,
                messages=messages[1:],
                max_tokens=8192,
                thinking={"type": "enabled", "budget_tokens": 4096},
                **kwargs,
            )
            thoughts = response.content[0].thinking
            print("CLAUDE 3.7 THOUGHTS:", thoughts)
            content = response.content[1].text
        else:
            response = self.llm_client.messages.create(
                system=messages[0]["content"][0]["text"],
                model=self.model,
                messages=messages[1:],
                max_tokens=max_new_tokens if max_new_tokens else 4096,
                temperature=temperature,
                **kwargs,
            )
            content = response.content[0].text
        
        total_tokens, cost = calculate_tokens_and_cost(response, self.provider, self.model)
        return content, total_tokens, cost


class LMMEngineGemini(LMMEngine):
    def __init__(
        self, base_url=None, api_key=None, model=None, rate_limit=-1, **kwargs
    ):
        assert model is not None, "model must be provided"
        self.model = model
        self.provider = "llm-gemini"

        api_key = api_key or os.getenv("GEMINI_API_KEY")
        if api_key is None:
            raise ValueError(
                "An API Key needs to be provided in either the api_key parameter or as an environment variable named GEMINI_API_KEY"
            )

        self.base_url = base_url or os.getenv("GEMINI_ENDPOINT_URL")
        if self.base_url is None:
            raise ValueError(
                "An endpoint URL needs to be provided in either the endpoint_url parameter or as an environment variable named GEMINI_ENDPOINT_URL"
            )

        self.api_key = api_key
        self.request_interval = 0 if rate_limit == -1 else 60.0 / rate_limit

        self.llm_client = OpenAI(base_url=self.base_url, api_key=self.api_key)

    @backoff.on_exception(
        backoff.expo, (APIConnectionError, APIError, RateLimitError), max_time=60
    )
    def generate(self, messages, temperature=0.0, max_new_tokens=None, **kwargs):
        """Generate the next message based on previous messages"""
        response = self.llm_client.chat.completions.create(
            model=self.model,
            messages=messages,
            max_tokens=max_new_tokens if max_new_tokens else 4096,
            temperature=temperature,
            # reasoning_effort="low",
            extra_body={
                'extra_body': {
                    "google": {
                        "thinking_config": {
                            "thinking_budget": 128,
                            "include_thoughts": True
                        }
                    }
                }
            },
            **kwargs,
        )
        
        content = response.choices[0].message.content
        total_tokens, cost = calculate_tokens_and_cost(response, self.provider, self.model)
        
        return content, total_tokens, cost



class LMMEngineOpenRouter(LMMEngine):
    def __init__(
        self, base_url=None, api_key=None, model=None, rate_limit=-1, **kwargs
    ):
        assert model is not None, "model must be provided"
        self.model = model
        self.provider = "llm-openrouter"

        api_key = api_key or os.getenv("OPENROUTER_API_KEY")
        if api_key is None:
            raise ValueError(
                "An API Key needs to be provided in either the api_key parameter or as an environment variable named OPENROUTER_API_KEY"
            )

        self.base_url = base_url or os.getenv("OPEN_ROUTER_ENDPOINT_URL")
        if self.base_url is None:
            raise ValueError(
                "An endpoint URL needs to be provided in either the endpoint_url parameter or as an environment variable named OPEN_ROUTER_ENDPOINT_URL"
            )

        self.api_key = api_key
        self.request_interval = 0 if rate_limit == -1 else 60.0 / rate_limit

        self.llm_client = OpenAI(base_url=self.base_url, api_key=self.api_key)

    @backoff.on_exception(
        backoff.expo, (APIConnectionError, APIError, RateLimitError), max_time=60
    )
    def generate(self, messages, temperature=0.0, max_new_tokens=None, **kwargs):
        """Generate the next message based on previous messages"""
        response = self.llm_client.chat.completions.create(
            model=self.model,
            messages=messages,
            max_tokens=max_new_tokens if max_new_tokens else 4096,
            temperature=temperature,
            **kwargs,
        )
        
        content = response.choices[0].message.content
        total_tokens, cost = calculate_tokens_and_cost(response, self.provider, self.model)
        
        return content, total_tokens, cost


class LMMEngineAzureOpenAI(LMMEngine):
    def __init__(
        self,
        base_url=None,
        api_key=None,
        azure_endpoint=None,
        model=None,
        api_version=None,
        rate_limit=-1,
        **kwargs
    ):
        assert model is not None, "model must be provided"
        self.model = model
        self.provider = "llm-azureopenai"

        assert api_version is not None, "api_version must be provided"
        self.api_version = api_version

        api_key = api_key or os.getenv("AZURE_OPENAI_API_KEY")
        if api_key is None:
            raise ValueError(
                "An API Key needs to be provided in either the api_key parameter or as an environment variable named AZURE_OPENAI_API_KEY"
            )

        self.api_key = api_key

        azure_endpoint = azure_endpoint or os.getenv("AZURE_OPENAI_ENDPOINT")
        if azure_endpoint is None:
            raise ValueError(
                "An Azure API endpoint needs to be provided in either the azure_endpoint parameter or as an environment variable named AZURE_OPENAI_ENDPOINT"
            )

        self.azure_endpoint = azure_endpoint
        self.request_interval = 0 if rate_limit == -1 else 60.0 / rate_limit

        self.llm_client = AzureOpenAI(
            azure_endpoint=self.azure_endpoint,
            api_key=self.api_key,
            api_version=self.api_version,
        )
        self.cost = 0.0

    # @backoff.on_exception(backoff.expo, (APIConnectionError, APIError, RateLimitError), max_tries=10)
    def generate(self, messages, temperature=0.0, max_new_tokens=None, **kwargs):
        """Generate the next message based on previous messages"""
        response = self.llm_client.chat.completions.create(
            model=self.model,
            messages=messages,
            max_tokens=max_new_tokens if max_new_tokens else 4096,
            temperature=temperature,
            **kwargs,
        )
        content = response.choices[0].message.content
        total_tokens, cost = calculate_tokens_and_cost(response, self.provider, self.model)
        return content, total_tokens, cost


class LMMEnginevLLM(LMMEngine):
    def __init__(
        self, base_url=None, api_key=None, model=None, rate_limit=-1, **kwargs
    ):
        assert model is not None, "model must be provided"
        self.model = model
        self.api_key = api_key
        self.provider = "llm-vllm"

        self.base_url = base_url or os.getenv("vLLM_ENDPOINT_URL")
        if self.base_url is None:
            raise ValueError(
                "An endpoint URL needs to be provided in either the endpoint_url parameter or as an environment variable named vLLM_ENDPOINT_URL"
            )

        self.request_interval = 0 if rate_limit == -1 else 60.0 / rate_limit

        self.llm_client = OpenAI(base_url=self.base_url, api_key=self.api_key)

    # @backoff.on_exception(backoff.expo, (APIConnectionError, APIError, RateLimitError), max_tries=10)
    # TODO: Default params chosen for the Qwen model
    def generate(
        self,
        messages,
        temperature=0.0,
        top_p=0.8,
        repetition_penalty=1.05,
        max_new_tokens=512,
        **kwargs
    ):
        """Generate the next message based on previous messages"""
        response = self.llm_client.chat.completions.create(
            model=self.model,
            messages=messages,
            max_tokens=max_new_tokens if max_new_tokens else 4096,
            temperature=temperature,
            top_p=top_p,
            extra_body={"repetition_penalty": repetition_penalty},
        )
        content = response.choices[0].message.content
        total_tokens, cost = calculate_tokens_and_cost(response, self.provider, self.model)
        return content, total_tokens, cost


class LMMEngineHuggingFace(LMMEngine):
    def __init__(self, base_url=None, api_key=None, rate_limit=-1, **kwargs):
        assert base_url is not None, "HuggingFace endpoint must be provided"
        self.base_url = base_url
        self.model = base_url.split('/')[-1] if base_url else "huggingface-tgi"
        self.provider = "llm-huggingface"

        api_key = api_key or os.getenv("HF_TOKEN")
        if api_key is None:
            raise ValueError(
                "A HuggingFace token needs to be provided in either the api_key parameter or as an environment variable named HF_TOKEN"
            )

        self.api_key = api_key
        self.request_interval = 0 if rate_limit == -1 else 60.0 / rate_limit

        self.llm_client = OpenAI(base_url=self.base_url, api_key=self.api_key)

    @backoff.on_exception(
        backoff.expo, (APIConnectionError, APIError, RateLimitError), max_time=60
    )
    def generate(self, messages, temperature=0.0, max_new_tokens=None, **kwargs):
        """Generate the next message based on previous messages"""
        response = self.llm_client.chat.completions.create(
            model="tgi",
            messages=messages,
            max_tokens=max_new_tokens if max_new_tokens else 4096,
            temperature=temperature,
            **kwargs,
        )
        
        content = response.choices[0].message.content
        total_tokens, cost = calculate_tokens_and_cost(response, self.provider, self.model)
        
        return content, total_tokens, cost


class LMMEngineDeepSeek(LMMEngine):
    def __init__(
        self, base_url=None, api_key=None, model=None, rate_limit=-1, **kwargs
    ):
        assert model is not None, "model must be provided"
        self.model = model
        self.provider = "llm-deepseek"

        api_key = api_key or os.getenv("DEEPSEEK_API_KEY")
        if api_key is None:
            raise ValueError(
                "An API Key needs to be provided in either the api_key parameter or as an environment variable named DEEPSEEK_API_KEY"
            )

        self.base_url = base_url or "https://api.deepseek.com"
        self.api_key = api_key
        self.request_interval = 0 if rate_limit == -1 else 60.0 / rate_limit

        self.llm_client = OpenAI(base_url=self.base_url, api_key=self.api_key)

    @backoff.on_exception(
        backoff.expo, (APIConnectionError, APIError, RateLimitError), max_time=60
    )
    def generate(self, messages, temperature=0.0, max_new_tokens=None, **kwargs):
        """Generate the next message based on previous messages"""
        response = self.llm_client.chat.completions.create(
            model=self.model,
            messages=messages,
            max_tokens=max_new_tokens if max_new_tokens else 4096,
            temperature=temperature,
            **kwargs,
        )
        
        content = response.choices[0].message.content
        total_tokens, cost = calculate_tokens_and_cost(response, self.provider, self.model)
        return content, total_tokens, cost


class LMMEngineZhipu(LMMEngine):
    def __init__(
        self, base_url=None, api_key=None, model=None, rate_limit=-1, **kwargs
    ):
        assert model is not None, "model must be provided"
        self.model = model
        self.provider = "llm-zhipu"

        api_key = api_key or os.getenv("ZHIPU_API_KEY")
        if api_key is None:
            raise ValueError(
                "An API Key needs to be provided in either the api_key parameter or as an environment variable named ZHIPU_API_KEY"
            )

        self.api_key = api_key
        self.request_interval = 0 if rate_limit == -1 else 60.0 / rate_limit

        # Use ZhipuAI client directly instead of OpenAI compatibility layer
        self.llm_client = ZhipuAI(api_key=self.api_key)

    @backoff.on_exception(
        backoff.expo, (APIConnectionError, APIError, RateLimitError), max_time=60
    )
    def generate(self, messages, temperature=0.0, max_new_tokens=None, **kwargs):
        """Generate the next message based on previous messages"""
        response = self.llm_client.chat.completions.create(
            model=self.model,
            messages=messages,
            max_tokens=max_new_tokens if max_new_tokens else 4096,
            temperature=temperature,
            **kwargs,
        )
        
        content = response.choices[0].message.content # type: ignore
        total_tokens, cost = calculate_tokens_and_cost(response, self.provider, self.model)
        return content, total_tokens, cost



class LMMEngineGroq(LMMEngine):
    def __init__(
        self, base_url=None, api_key=None, model=None, rate_limit=-1, **kwargs
    ):
        assert model is not None, "model must be provided"
        self.model = model
        self.provider = "llm-groq"

        api_key = api_key or os.getenv("GROQ_API_KEY")
        if api_key is None:
            raise ValueError(
                "An API Key needs to be provided in either the api_key parameter or as an environment variable named GROQ_API_KEY"
            )

        self.api_key = api_key
        self.request_interval = 0 if rate_limit == -1 else 60.0 / rate_limit

        # Use Groq client directly
        self.llm_client = Groq(api_key=self.api_key)

    @backoff.on_exception(
        backoff.expo, (APIConnectionError, APIError, RateLimitError), max_time=60
    )
    def generate(self, messages, temperature=0.0, max_new_tokens=None, **kwargs):
        """Generate the next message based on previous messages"""
        response = self.llm_client.chat.completions.create(
            model=self.model,
            messages=messages,
            max_tokens=max_new_tokens if max_new_tokens else 4096,
            temperature=temperature,
            **kwargs,
        )
        
        content = response.choices[0].message.content
        total_tokens, cost = calculate_tokens_and_cost(response, self.provider, self.model)
        return content, total_tokens, cost


class LMMEngineSiliconflow(LMMEngine):
    def __init__(
        self, base_url=None, api_key=None, model=None, rate_limit=-1, **kwargs
    ):
        assert model is not None, "model must be provided"
        self.model = model
        self.provider = "llm-siliconflow"

        api_key = api_key or os.getenv("SILICONFLOW_API_KEY")
        if api_key is None:
            raise ValueError(
                "An API Key needs to be provided in either the api_key parameter or as an environment variable named SILICONFLOW_API_KEY"
            )

        self.base_url = base_url or "https://api.siliconflow.cn/v1"
        self.api_key = api_key
        self.request_interval = 0 if rate_limit == -1 else 60.0 / rate_limit

        self.llm_client = OpenAI(base_url=self.base_url, api_key=self.api_key)

    @backoff.on_exception(
        backoff.expo, (APIConnectionError, APIError, RateLimitError), max_time=60
    )
    def generate(self, messages, temperature=0.0, max_new_tokens=None, **kwargs):
        """Generate the next message based on previous messages"""
        response = self.llm_client.chat.completions.create(
            model=self.model,
            messages=messages,
            max_tokens=max_new_tokens if max_new_tokens else 4096,
            temperature=temperature,
            **kwargs,
        )
        
        content = response.choices[0].message.content
        total_tokens, cost = calculate_tokens_and_cost(response, self.provider, self.model)
        return content, total_tokens, cost


class LMMEngineMonica(LMMEngine):
    def __init__(
        self, base_url=None, api_key=None, model=None, rate_limit=-1, **kwargs
    ):
        assert model is not None, "model must be provided"
        self.model = model
        self.provider = "llm-monica"

        api_key = api_key or os.getenv("MONICA_API_KEY")
        if api_key is None:
            raise ValueError(
                "An API Key needs to be provided in either the api_key parameter or as an environment variable named MONICA_API_KEY"
            )

        self.base_url = base_url or "https://openapi.monica.im/v1"
        self.api_key = api_key
        self.request_interval = 0 if rate_limit == -1 else 60.0 / rate_limit

        self.llm_client = OpenAI(base_url=self.base_url, api_key=self.api_key)

    @backoff.on_exception(
        backoff.expo, (APIConnectionError, APIError, RateLimitError), max_time=60
    )
    def generate(self, messages, temperature=0.0, max_new_tokens=None, **kwargs):
        """Generate the next message based on previous messages"""
        response = self.llm_client.chat.completions.create(
            model=self.model,
            messages=messages,
            max_tokens=max_new_tokens if max_new_tokens else 4096,
            temperature=temperature,
            **kwargs,
        )
        
        content = response.choices[0].message.content
        total_tokens, cost = calculate_tokens_and_cost(response, self.provider, self.model)
        return content, total_tokens, cost


class LMMEngineAWSBedrock(LMMEngine):
    def __init__(
        self,
        aws_access_key=None,
        aws_secret_key=None,
        aws_region=None,
        model=None,
        rate_limit=-1,
        **kwargs
    ):
        assert model is not None, "model must be provided"
        self.model = model
        self.provider = "llm-bedrock"

        # Claude model mapping for AWS Bedrock
        self.claude_model_map = {
            "claude-opus-4": "anthropic.claude-opus-4-20250514-v1:0",
            "claude-sonnet-4": "anthropic.claude-sonnet-4-20250514-v1:0",
            "claude-3-7-sonnet": "anthropic.claude-3-7-sonnet-20250219-v1:0",
            "claude-3-5-sonnet": "anthropic.claude-3-5-sonnet-20241022-v2:0",
            "claude-3-5-sonnet-20241022": "anthropic.claude-3-5-sonnet-20241022-v2:0",
            "claude-3-5-sonnet-20240620": "anthropic.claude-3-5-sonnet-20240620-v1:0",
            "claude-3-5-haiku": "anthropic.claude-3-5-haiku-20241022-v1:0",
            "claude-3-haiku": "anthropic.claude-3-haiku-20240307-v1:0",
            "claude-3-sonnet": "anthropic.claude-3-sonnet-20240229-v1:0",
            "claude-3-opus": "anthropic.claude-3-opus-20240229-v1:0",
        }

        # Get the actual Bedrock model ID
        self.bedrock_model_id = self.claude_model_map.get(model, model)

        # AWS credentials
        aws_access_key = aws_access_key or os.getenv("AWS_ACCESS_KEY_ID")
        aws_secret_key = aws_secret_key or os.getenv("AWS_SECRET_ACCESS_KEY")
        aws_region = aws_region or os.getenv("AWS_DEFAULT_REGION") or "us-west-2"

        if aws_access_key is None:
            raise ValueError(
                "AWS Access Key needs to be provided in either the aws_access_key parameter or as an environment variable named AWS_ACCESS_KEY_ID"
            )
        if aws_secret_key is None:
            raise ValueError(
                "AWS Secret Key needs to be provided in either the aws_secret_key parameter or as an environment variable named AWS_SECRET_ACCESS_KEY"
            )

        self.aws_region = aws_region
        self.request_interval = 0 if rate_limit == -1 else 60.0 / rate_limit

        # Initialize Bedrock client
        self.bedrock_client = boto3.client(
            service_name="bedrock-runtime",
            region_name=aws_region,
            aws_access_key_id=aws_access_key,
            aws_secret_access_key=aws_secret_key
        )

    @backoff.on_exception(
        backoff.expo, (APIConnectionError, APIError, RateLimitError), max_time=60
    )
    def generate(self, messages, temperature=0.0, max_new_tokens=None, **kwargs):
        """Generate the next message based on previous messages"""

        # Convert messages to Bedrock format
        # Extract system message if present
        system_message = None
        user_messages = []

        for message in messages:
            if message["role"] == "system":
                if isinstance(message["content"], list):
                    system_message = message["content"][0]["text"]
                else:
                    system_message = message["content"]
            else:
                # Handle both list and string content formats
                if isinstance(message["content"], list):
                    content = message["content"][0]["text"] if message["content"] else ""
                else:
                    content = message["content"]

                user_messages.append({
                    "role": message["role"],
                    "content": content
                })

        # Prepare the body for Bedrock
        body = {
            "max_tokens": max_new_tokens if max_new_tokens else 4096,
            "messages": user_messages,
            "anthropic_version": "bedrock-2023-05-31"
        }

        if temperature > 0:
            body["temperature"] = temperature

        if system_message:
            body["system"] = system_message

        try:
            response = self.bedrock_client.invoke_model(
                body=json.dumps(body),
                modelId=self.bedrock_model_id
            )

            response_body = json.loads(response.get("body").read())

            if "content" in response_body and response_body["content"]:
                content = response_body["content"][0]["text"]
            else:
                raise ValueError("No content in response")
            
            total_tokens, cost = calculate_tokens_and_cost(response_body, self.provider, self.model)
            return content, total_tokens, cost

        except Exception as e:
            print(f"AWS Bedrock error: {e}")
            raise

# ==================== Embedding ====================

class OpenAIEmbeddingEngine(LMMEngine):
    def __init__(
        self,
        embedding_model: str = "text-embedding-3-small",
        api_key=None,
        **kwargs
    ):
        """Init an OpenAI Embedding engine

        Args:
            embedding_model (str, optional): Model name. Defaults to "text-embedding-3-small".
            api_key (_type_, optional): Auth key from OpenAI. Defaults to None.
        """
        self.model = embedding_model
        self.provider = "embedding-openai"

        api_key = api_key or os.getenv("OPENAI_API_KEY")
        if api_key is None:
            raise ValueError(
                "An API Key needs to be provided in either the api_key parameter or as an environment variable named OPENAI_API_KEY"
            )
        self.api_key = api_key

    @backoff.on_exception(
        backoff.expo,
        (
            APIError,
            RateLimitError,
            APIConnectionError,
        ),
    )
    def get_embeddings(self, text: str) -> Tuple[np.ndarray, List[int], float]:
        client = OpenAI(api_key=self.api_key)
        response = client.embeddings.create(model=self.model, input=text)
        
        embeddings = np.array([data.embedding for data in response.data])
        total_tokens, cost = calculate_tokens_and_cost(response, self.provider, self.model)
        
        return embeddings, total_tokens, cost



class GeminiEmbeddingEngine(LMMEngine):
    def __init__(
        self,
        embedding_model: str = "text-embedding-004",
        api_key=None,
        **kwargs
    ):
        """Init an Gemini Embedding engine

        Args:
            embedding_model (str, optional): Model name. Defaults to "text-embedding-004".
            api_key (_type_, optional): Auth key from Gemini. Defaults to None.
        """
        self.model = embedding_model
        self.provider = "embedding-gemini"

        api_key = api_key or os.getenv("GEMINI_API_KEY")
        if api_key is None:
            raise ValueError(
                "An API Key needs to be provided in either the api_key parameter or as an environment variable named GEMINI_API_KEY"
            )
        self.api_key = api_key

    @backoff.on_exception(
        backoff.expo,
        (
            APIError,
            RateLimitError,
            APIConnectionError,
        ),
    )
    def get_embeddings(self, text: str) -> Tuple[np.ndarray, List[int], float]:
        client = genai.Client(api_key=self.api_key)

        result = client.models.embed_content(
            model=self.model,
            contents=text,
            config=types.EmbedContentConfig(task_type="SEMANTIC_SIMILARITY"),
        )

        embeddings = np.array([i.values for i in result.embeddings]) # type: ignore
        total_tokens, cost = calculate_tokens_and_cost(result, self.provider, self.model)
        
        return embeddings, total_tokens, cost



class AzureOpenAIEmbeddingEngine(LMMEngine):
    def __init__(
        self,
        embedding_model: str = "text-embedding-3-small",
        api_key=None,
        api_version=None,
        endpoint_url=None,
        **kwargs
    ):
        """Init an Azure OpenAI Embedding engine

        Args:
            embedding_model (str, optional): Model name. Defaults to "text-embedding-3-small".
            api_key (_type_, optional): Auth key from Azure OpenAI. Defaults to None.
            api_version (_type_, optional): API version. Defaults to None.
            endpoint_url (_type_, optional): Endpoint URL. Defaults to None.
        """
        self.model = embedding_model
        self.provider = "embedding-azureopenai"

        api_key = api_key or os.getenv("AZURE_OPENAI_API_KEY")
        if api_key is None:
            raise ValueError(
                "An API Key needs to be provided in either the api_key parameter or as an environment variable named AZURE_OPENAI_API_KEY"
            )
        self.api_key = api_key

        api_version = api_version or os.getenv("OPENAI_API_VERSION")
        if api_version is None:
            raise ValueError(
                "An API Version needs to be provided in either the api_version parameter or as an environment variable named OPENAI_API_VERSION"
            )
        self.api_version = api_version

        endpoint_url = endpoint_url or os.getenv("AZURE_OPENAI_ENDPOINT")
        if endpoint_url is None:
            raise ValueError(
                "An Endpoint URL needs to be provided in either the endpoint_url parameter or as an environment variable named AZURE_OPENAI_ENDPOINT"
            )
        self.endpoint_url = endpoint_url

    @backoff.on_exception(
        backoff.expo,
        (
            APIError,
            RateLimitError,
            APIConnectionError,
        ),
    )
    def get_embeddings(self, text: str) -> Tuple[np.ndarray, List[int], float]:
        client = AzureOpenAI(
            api_key=self.api_key,
            api_version=self.api_version,
            azure_endpoint=self.endpoint_url,
        )
        response = client.embeddings.create(input=text, model=self.model)
        
        embeddings = np.array([data.embedding for data in response.data])
        total_tokens, cost = calculate_tokens_and_cost(response, self.provider, self.model)
        
        return embeddings, total_tokens, cost


class DashScopeEmbeddingEngine(LMMEngine):
    def __init__(
        self,
        embedding_model: str = "text-embedding-v4",
        api_key=None,
        dimensions: int = 1024,
        **kwargs
    ):
        """Init a DashScope (阿里云百炼) Embedding engine

        Args:
            embedding_model (str, optional): Model name. Defaults to "text-embedding-v4".
            api_key (_type_, optional): Auth key from DashScope. Defaults to None.
            dimensions (int, optional): Embedding dimensions. Defaults to 1024.
        """
        self.model = embedding_model
        self.dimensions = dimensions
        self.provider = "embedding-qwen"

        api_key = api_key or os.getenv("DASHSCOPE_API_KEY")
        if api_key is None:
            raise ValueError(
                "An API Key needs to be provided in either the api_key parameter or as an environment variable named DASHSCOPE_API_KEY"
            )
        self.api_key = api_key

        # Initialize OpenAI client with DashScope base URL
        self.client = OpenAI(
            api_key=self.api_key,
            base_url="https://dashscope.aliyuncs.com/compatible-mode/v1"
        )

    @backoff.on_exception(
        backoff.expo,
        (
                APIError,
                RateLimitError,
                APIConnectionError,
        ),
    )
    def get_embeddings(self, text: str) -> Tuple[np.ndarray, List[int], float]:
        response = self.client.embeddings.create(
            model=self.model,
            input=text,
            dimensions=self.dimensions,
            encoding_format="float"
        )
        
        embeddings = np.array([data.embedding for data in response.data])
        total_tokens, cost = calculate_tokens_and_cost(response, self.provider, self.model)
        
        return embeddings, total_tokens, cost



class DoubaoEmbeddingEngine(LMMEngine):
    def __init__(
        self,
        embedding_model: str = "doubao-embedding-256",
        api_key=None,
        **kwargs
    ):
        """Init a Doubao (字节跳动豆包) Embedding engine

        Args:
            embedding_model (str, optional): Model name. Defaults to "doubao-embedding-256".
            api_key (_type_, optional): Auth key from Doubao. Defaults to None.
        """
        self.model = embedding_model
        self.provider = "embedding-doubao"

        api_key = api_key or os.getenv("ARK_API_KEY")
        if api_key is None:
            raise ValueError(
                "An API Key needs to be provided in either the api_key parameter or as an environment variable named ARK_API_KEY"
            )
        self.api_key = api_key
        self.base_url = "https://ark.cn-beijing.volces.com/api/v3"

        # Use OpenAI-compatible client for text embeddings
        self.client = OpenAI(
            api_key=self.api_key,
            base_url=self.base_url
        )

    @backoff.on_exception(
        backoff.expo,
        (
                APIError,
                RateLimitError,
                APIConnectionError,
        ),
    )
    def get_embeddings(self, text: str) -> Tuple[np.ndarray, List[int], float]:
        response = self.client.embeddings.create(
            model=self.model,
            input=text,
            encoding_format="float"
        )
        
        embeddings = np.array([data.embedding for data in response.data])
        total_tokens, cost = calculate_tokens_and_cost(response, self.provider, self.model)
        
        return embeddings, total_tokens, cost


class JinaEmbeddingEngine(LMMEngine):
    def __init__(
        self,
        embedding_model: str = "jina-embeddings-v4",
        api_key=None,
        task: str = "retrieval.query",
        **kwargs
    ):
        """Init a Jina AI Embedding engine

        Args:
            embedding_model (str, optional): Model name. Defaults to "jina-embeddings-v4".
            api_key (_type_, optional): Auth key from Jina AI. Defaults to None.
            task (str, optional): Task type. Options: "retrieval.query", "retrieval.passage", "text-matching". Defaults to "retrieval.query".
        """
        self.model = embedding_model
        self.task = task
        self.provider = "embedding-jina" 

        api_key = api_key or os.getenv("JINA_API_KEY")
        if api_key is None:
            raise ValueError(
                "An API Key needs to be provided in either the api_key parameter or as an environment variable named JINA_API_KEY"
            )
        self.api_key = api_key
        self.base_url = "https://api.jina.ai/v1"

    @backoff.on_exception(
        backoff.expo,
        (
                APIError,
                RateLimitError,
                APIConnectionError,
        ),
    )
    def get_embeddings(self, text: str) -> Tuple[np.ndarray, List[int], float]:
        import requests

        headers = {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {self.api_key}"
        }

        data = {
            "model": self.model,
            "task": self.task,
            "input": [
                {
                    "text": text
                }
            ]
        }

        response = requests.post(
            f"{self.base_url}/embeddings",
            headers=headers,
            json=data
        )

        if response.status_code != 200:
            raise Exception(f"Jina AI API error: {response.text}")

        result = response.json()
        embeddings = np.array([data["embedding"] for data in result["data"]])
        
        total_tokens, cost = calculate_tokens_and_cost(result, self.provider, self.model)
    
        return embeddings, total_tokens, cost


# ==================== webSearch ====================
class SearchEngine:
    """Base class for search engines"""
    pass

class BochaAISearchEngine(SearchEngine):
    def __init__(
            self,
            api_key: str|None = None,
            base_url: str = "https://api.bochaai.com/v1",
            rate_limit: int = -1,
            **kwargs
    ):
        """Init a Bocha AI Search engine

        Args:
            api_key (str, optional): Auth key from Bocha AI. Defaults to None.
            base_url (str, optional): Base URL for the API. Defaults to "https://api.bochaai.com/v1".
            rate_limit (int, optional): Rate limit per minute. Defaults to -1 (no limit).
        """
        api_key = api_key or os.getenv("BOCHA_API_KEY")
        if api_key is None:
            raise ValueError(
                "An API Key needs to be provided in either the api_key parameter or as an environment variable named BOCHA_API_KEY"
            )

        self.api_key = api_key
        self.base_url = base_url
        self.endpoint = f"{base_url}/ai-search"
        self.request_interval = 0 if rate_limit == -1 else 60.0 / rate_limit

    @backoff.on_exception(
        backoff.expo,
        (
                APIConnectionError,
                APIError,
                RateLimitError,
                requests.exceptions.RequestException,
        ),
        max_time=60
    )
    def search(
            self,
            query: str,
            freshness: str = "noLimit",
            answer: bool = True,
            stream: bool = False,
            **kwargs
    ) -> Union[Dict[str, Any], Any]:
        """Search with AI and return intelligent answer

        Args:
            query (str): Search query
            freshness (str, optional): Freshness filter. Defaults to "noLimit".
            answer (bool, optional): Whether to return answer. Defaults to True.
            stream (bool, optional): Whether to stream response. Defaults to False.

        Returns:
            Union[Dict[str, Any], Any]: AI search results with sources and answer
        """
        headers = {
            'Authorization': f'Bearer {self.api_key}',
            'Content-Type': 'application/json'
        }

        payload = {
            "query": query,
            "freshness": freshness,
            "answer": answer,
            "stream": stream,
            **kwargs
        }

        if stream:
            result = self._stream_search(headers, payload)
            return result, [0, 0, 0], 0.06
        else:
            result = self._regular_search(headers, payload)
            return result, [0, 0, 0], 0.06


    def _regular_search(self, headers: Dict[str, str], payload: Dict[str, Any]) -> Dict[str, Any]:
        """Regular non-streaming search"""
        response = requests.post(
            self.endpoint,
            headers=headers,
            json=payload
        )

        if response.status_code != 200:
            raise APIError(f"Bocha AI Search API error: {response.text}") # type: ignore

        return response.json()

    def _stream_search(self, headers: Dict[str, str], payload: Dict[str, Any]):
        """Streaming search response"""
        response = requests.post(
            self.endpoint,
            headers=headers,
            json=payload,
            stream=True
        )

        if response.status_code != 200:
            raise APIError(f"Bocha AI Search API error: {response.text}") # type: ignore

        for line in response.iter_lines():
            if line:
                line = line.decode('utf-8')
                if line.startswith('data:'):
                    data = line[5:].strip()
                    if data and data != '{"event":"done"}':
                        try:
                            yield json.loads(data)
                        except json.JSONDecodeError:
                            continue

    def get_answer(self, query: str, **kwargs) -> Tuple[str, int, float]:
        """Get AI generated answer only"""
        result, _, remaining_balance = self.search(query, answer=True, **kwargs)

        # Extract answer from messages
        messages = result.get("messages", []) # type: ignore
        answer = ""
        for message in messages:
            if message.get("type") == "answer":
                answer = message.get("content", "")
                break

        return answer, [0,0,0], remaining_balance # type: ignore


    def get_sources(self, query: str, **kwargs) -> List[Dict[str, Any]]:
        """Get source materials only"""
        result, _, remaining_balance = self.search(query, **kwargs)

        # Extract sources from messages
        sources = []
        messages = result.get("messages", []) # type: ignore
        for message in messages:
            if message.get("type") == "source":
                content_type = message.get("content_type", "")
                if content_type in ["webpage", "image", "video", "baike_pro", "medical_common"]:
                    sources.append({
                        "type": content_type,
                        "content": json.loads(message.get("content", "{}"))
                    })

        return sources, 0, remaining_balance # type: ignore


    def get_follow_up_questions(self, query: str, **kwargs) -> List[str]:
        """Get follow-up questions"""
        result, _, remaining_balance = self.search(query, **kwargs)

        # Extract follow-up questions from messages
        follow_ups = []
        messages = result.get("messages", []) # type: ignore
        for message in messages:
            if message.get("type") == "follow_up":
                follow_ups.append(message.get("content", ""))

        return follow_ups, 0, remaining_balance # type: ignore


class ExaResearchEngine(SearchEngine):
    def __init__(
            self,
            api_key: str|None = None,
            base_url: str = "https://api.exa.ai",
            rate_limit: int = -1,
            **kwargs
    ):
        """Init an Exa Research engine

        Args:
            api_key (str, optional): Auth key from Exa AI. Defaults to None.
            base_url (str, optional): Base URL for the API. Defaults to "https://api.exa.ai".
            rate_limit (int, optional): Rate limit per minute. Defaults to -1 (no limit).
        """
        api_key = api_key or os.getenv("EXA_API_KEY")
        if api_key is None:
            raise ValueError(
                "An API Key needs to be provided in either the api_key parameter or as an environment variable named EXA_API_KEY"
            )

        self.api_key = api_key
        self.base_url = base_url
        self.request_interval = 0 if rate_limit == -1 else 60.0 / rate_limit

        # Initialize OpenAI-compatible client for chat completions
        self.chat_client = OpenAI(
            base_url=base_url,
            api_key=api_key
        )

        # Initialize Exa client for research tasks
        try:
            from exa_py import Exa
            self.exa_client = Exa(api_key=api_key)
        except ImportError:
            self.exa_client = None
            print("Warning: exa_py not installed. Research tasks will not be available.")

    @backoff.on_exception(
        backoff.expo,
        (
                APIConnectionError,
                APIError,
                RateLimitError,
        ),
        max_time=60
    )
    def search(self, query: str, **kwargs):
        """Standard Exa search with direct cost from API

        Args:
            query (str): Search query
            **kwargs: Additional search parameters

        Returns:
            tuple: (result, tokens, cost) where cost is actual API cost
        """
        headers = {
            'x-api-key': self.api_key,
            'Content-Type': 'application/json'
        }
        
        payload = {
            "query": query,
            **kwargs
        }
        
        response = requests.post(
            f"{self.base_url}/search",
            headers=headers,
            json=payload
        )
        
        if response.status_code != 200:
            raise APIError(f"Exa Search API error: {response.text}") # type: ignore
        
        result = response.json()
        
        cost = 0.0
        if "costDollars" in result:
            cost = result["costDollars"].get("total", 0.0)
        
        return result, [0, 0, 0], cost
    
    def chat_research(
            self,
            query: str,
            model: str = "exa",
            stream: bool = False,
            **kwargs
    ) -> Union[str, Any]:
        """Research using chat completions interface

        Args:
            query (str): Research query
            model (str, optional): Model name. Defaults to "exa".
            stream (bool, optional): Whether to stream response. Defaults to False.

        Returns:
            Union[str, Any]: Research result or stream
        """
        messages = [
            {"role": "user", "content": query}
        ]

        if stream:
            completion = self.chat_client.chat.completions.create(
                model=model,
                messages=messages, # type: ignore
                stream=True,
                **kwargs
            )
            return completion
        else:
            completion = self.chat_client.chat.completions.create(
                model=model,
                messages=messages, # type: ignore
                **kwargs
            )
            result = completion.choices[0].message.content # type: ignore
            return result,[0,0,0],0.005
